"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import NULL_VALUE


class ResultsetRowPacket(Packet):
    """
    Implementation based on:
    https://dev.mysql.com/doc/internals/en/com-query-response.html#packet-ProtocolText::ResultsetRow
    https://mariadb.com/kb/en/resultset-row/
    """

    def setup(self):
        data = self._kwargs.get("data", {})
        self.value = []
        for val in data:
            if val is None:
                self.value.append(NULL_VALUE)
            elif isinstance(val, bytes):
                self.value.append(Datum("byte<lenenc>", val))
            else:
                self.value.append(Datum("string<lenenc>", str(val)))

    @property
    def body(self):
        string = b""
        for x in self.value:
            if x is NULL_VALUE:
                string += x
            else:
                string += x.toStringPacket()

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint

        pprint.pprint(str(ResultsetRowPacket().get_packet_string()))


if __name__ == "__main__":
    ResultsetRowPacket.test()



class RESPONSE_TYPE:
    __slots__ = ()
    OK = 'ok'
    TABLE = 'table'
    ERROR = 'error'
    COLUMNS_TABLE = 'columns_table'  # for queries to information_schema.columns


RESPONSE_TYPE = RESPONSE_TYPE()



from sklearn.metrics import r2_score
import dill
import pandas as pd
import tempfile
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.time_series_utils import (
    transform_to_nixtla_df,
    get_results_from_nixtla_df,
    infer_frequency,
    get_model_accuracy_dict,
    get_hierarchy_from_df,
    reconcile_forecasts
)
from neuralforecast import NeuralForecast
from neuralforecast.models import NHITS
from neuralforecast.auto import AutoNHITS
from ray.tune.search.hyperopt import HyperOptSearch

# hierarchicalforecast is an optional dependency
try:
    from hierarchicalforecast.core import HierarchicalReconciliation
except ImportError:
    HierarchicalReconciliation = None


class NeuralForecastHandler(BaseMLEngine):
    """Integration with the Nixtla NeuralForecast library for
    time series forecasting with neural networks.
    """

    name = "neuralforecast"

    def create(self, target, df, args={}):
        """Create the NeuralForecast Handler.

        Requires specifying the target column to predict and time series arguments for
        prediction horizon, time column (order by) and grouping column(s).

        Saves model params to desk, which are called later in the predict() method.
        """
        time_settings = args["timeseries_settings"]
        using_args = args["using"]
        assert time_settings["is_timeseries"], "Specify time series settings in your query"
        # store model args and time series settings in the model folder
        model_args = {}
        model_args["target"] = target
        model_args["horizon"] = time_settings["horizon"]
        model_args["order_by"] = time_settings["order_by"]
        model_args["group_by"] = time_settings["group_by"]
        model_args["frequency"] = (
            using_args["frequency"] if "frequency" in using_args else infer_frequency(df, time_settings["order_by"])
        )
        model_args["exog_vars"] = using_args["exogenous_vars"] if "exogenous_vars" in using_args else []
        model_args["max_steps"] = using_args.get('max_steps', 20)
        model_args["val_check_steps"] = using_args.get('val_check_steps', 10)
        model_args["n_auto_trials"] = using_args.get('n_auto_trials', 0)
        model_args["model_folder"] = tempfile.mkdtemp()

        # Deal with hierarchy
        model_args["hierarchy"] = using_args["hierarchy"] if "hierarchy" in using_args else False
        if model_args["hierarchy"] and HierarchicalReconciliation is not None:
            training_df, hier_df, hier_dict = get_hierarchy_from_df(df, model_args)
            self.model_storage.file_set("hier_dict", dill.dumps(hier_dict))
            self.model_storage.file_set("hier_df", dill.dumps(hier_df))
            self.model_storage.file_set("training_df", dill.dumps(training_df))
        else:
            training_df = transform_to_nixtla_df(df, model_args, model_args["exog_vars"])

        # Train model
        if model_args["n_auto_trials"]:
            model = AutoNHITS(time_settings["horizon"], gpus=0, num_samples=model_args["n_auto_trials"], search_alg=HyperOptSearch())
        else:
            # faster implementation without auto parameter tuning
            model = NHITS(time_settings["horizon"], time_settings["window"], hist_exog_list=model_args["exog_vars"], max_steps=model_args["max_steps"])
        neural = NeuralForecast(models=[model], freq=model_args["frequency"])

        if model_args.get('crossval', False):
            results_df = neural.cross_validation(training_df)
            model_args["accuracies"] = get_model_accuracy_dict(results_df, r2_score)
        else:
            neural.fit(training_df)

        # persist changes to handler folder
        neural.save(model_args["model_folder"], overwrite=True)
        self.model_storage.json_set("model_args", model_args)

    def predict(self, df, args={}):
        """Makes forecasts with the NeuralForecast Handler.

        NeuralForecast is setup to predict for all groups, so it won't handle
        a dataframe that's been filtered to one group very well. Instead, we make
        the prediction for all groups then take care of the filtering after the
        forecasting. Prediction is nearly instant.
        """
        # Load model arguments
        model_args = self.model_storage.json_get("model_args")

        prediction_df = transform_to_nixtla_df(df, model_args)
        groups_to_keep = prediction_df["unique_id"].unique()

        neural = NeuralForecast.load(model_args["model_folder"])
        forecast_df = neural.predict()
        if model_args["hierarchy"] and HierarchicalReconciliation is not None:
            training_df = dill.loads(self.model_storage.file_get("training_df"))
            hier_df = dill.loads(self.model_storage.file_get("hier_df"))
            hier_dict = dill.loads(self.model_storage.file_get("hier_dict"))
            reconciled_df = reconcile_forecasts(training_df, forecast_df, hier_df, hier_dict)
            results_df = reconciled_df[reconciled_df.index.isin(groups_to_keep)]
        else:
            results_df = forecast_df[forecast_df.index.isin(groups_to_keep)].rename({
                "y": model_args["target"],  # auto mode
                "NHITS": model_args["target"],  # non-auto mode
            }, axis=1)
        return get_results_from_nixtla_df(results_df, model_args)

    def describe(self, attribute=None):
        model_args = self.model_storage.json_get("model_args")

        if attribute == "model":
            return pd.DataFrame({k: [model_args[k]] for k in ["model_name", "frequency", "hierarchy"]})

        elif attribute == "features":
            return pd.DataFrame(
                {"ds": [model_args["order_by"]], "y": model_args["target"], "unique_id": [model_args["group_by"]], "exog_vars": [model_args["exog_vars"]]}
            )

        elif attribute == 'info':
            outputs = model_args["target"]
            inputs = [model_args["target"], model_args["order_by"], model_args["group_by"]] + model_args["exog_vars"]
            accuracies = [(model, acc) for model, acc in model_args.get("accuracies", {}).items()]
            return pd.DataFrame({"accuracies": [accuracies], "outputs": outputs, "inputs": [inputs]})

        else:
            tables = ['info', 'features', 'model']
            return pd.DataFrame(tables, columns=['tables'])




from langchain_core.embeddings import Embeddings
from langchain_community.vectorstores import Chroma, PGVector
from langchain_core.vectorstores import VectorStore

from pydantic import BaseModel

from mindsdb.integrations.utilities.rag.settings import VectorStoreType, VectorStoreConfig
from mindsdb.integrations.utilities.rag.loaders.vector_store_loader.MDBVectorStore import MDBVectorStore
from mindsdb.integrations.utilities.rag.loaders.vector_store_loader.pgvector import PGVectorMDB
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class VectorStoreLoader(BaseModel):
    embedding_model: Embeddings
    vector_store: VectorStore = None
    config: VectorStoreConfig = None

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"
        validate_assignment = True

    def load(self) -> VectorStore:
        """
        Loads the vector store based on the provided config and embeddings model
        :return:
        """
        if self.config.is_sparse is not None and self.config.vector_size is not None and self.config.kb_table is not None:
            # Only use PGVector store for sparse vectors.
            db_handler = self.config.kb_table.get_vector_db()
            db_args = db_handler.connection_args
            # Assume we are always using PGVector & psycopg2.
            connection_str = f"postgresql+psycopg2://{db_args.get('user')}:{db_args.get('password')}@{db_args.get('host')}:{db_args.get('port')}/{db_args.get('dbname', db_args.get('database'))}"

            return PGVectorMDB(
                connection_string=connection_str,
                collection_name=self.config.kb_table._kb.vector_database_table,
                embedding_function=self.embedding_model,
                is_sparse=self.config.is_sparse,
                vector_size=self.config.vector_size
            )
        return MDBVectorStore(kb_table=self.config.kb_table)


class VectorStoreFactory:
    @staticmethod
    def create(embedding_model: Embeddings, config: VectorStoreConfig):

        if config.vector_store_type == VectorStoreType.CHROMA:
            return VectorStoreFactory._load_chromadb_store(embedding_model, config)
        elif config.vector_store_type == VectorStoreType.PGVECTOR:
            return VectorStoreFactory._load_pgvector_store(embedding_model, config)
        else:
            raise ValueError(f"Invalid vector store type, must be one either {VectorStoreType.__members__.keys()}")

    @staticmethod
    def _load_chromadb_store(embedding_model: Embeddings, settings) -> Chroma:
        return Chroma(
            persist_directory=settings.persist_directory,
            collection_name=settings.collection_name,
            embedding_function=embedding_model,
        )

    @staticmethod
    def _load_pgvector_store(embedding_model: Embeddings, settings) -> PGVector:
        from .pgvector import PGVectorMDB
        return PGVectorMDB(
            connection_string=settings.connection_string,
            collection_name=settings.collection_name,
            embedding_function=embedding_model,
            is_sparse=settings.is_sparse,
            vector_size=settings.vector_size
        )



"""
Wrapper around MindsDB's executor and integration controller following the implementation of the original
langchain.sql_database.SQLDatabase class to partly replicate its behavior.
"""

import traceback
from typing import Any, Iterable, List, Optional

from mindsdb.utilities import log
from langchain_community.utilities import SQLDatabase
from mindsdb.interfaces.skills.sql_agent import SQLAgent

logger = log.getLogger(__name__)


def extract_essential(input: str) -> str:
    """Sometimes LLM include to input unnecessary data. We can't control stochastic nature of LLM, so we need to
    'clean' input somehow. LLM prompt contains instruction to enclose input between '$START$' and '$STOP$'.
    """
    if "$START$" in input:
        input = input.partition("$START$")[-1]
    if "$STOP$" in input:
        input = input.partition("$STOP$")[0]
    return input.strip(" ")


class MindsDBSQL(SQLDatabase):
    @staticmethod
    def custom_init(sql_agent: "SQLAgent") -> "MindsDBSQL":
        instance = MindsDBSQL()
        instance._sql_agent = sql_agent
        return instance

    """ Can't modify signature, as LangChain does a Pydantic check."""

    def __init__(
        self,
        engine: Optional[Any] = None,
        schema: Optional[str] = None,
        metadata: Optional[Any] = None,
        ignore_tables: Optional[List[str]] = None,
        include_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        indexes_in_table_info: bool = False,
        custom_table_info: Optional[dict] = None,
        view_support: bool = True,
        max_string_length: int = 300,
        lazy_table_reflection: bool = False,
    ):
        pass

    @property
    def dialect(self) -> str:
        return "mindsdb"

    @property
    def table_info(self) -> str:
        """Information about all tables in the database."""
        return self._sql_agent.get_table_info()

    @property
    def kb_info(self) -> str:
        """Information about all knowledge bases in the database."""
        return self._sql_agent.get_knowledge_base_info()

    def get_usable_table_names(self) -> Iterable[str]:
        """Information about all tables in the database."""
        try:
            return self._sql_agent.get_usable_table_names()
        except Exception as e:
            # If there's an error accessing tables through the integration, handle it gracefully
            logger.warning(f"Error getting usable table names: {str(e)}")
            # Return an empty list instead of raising an exception
            return []

    def get_table_info_no_throw(self, table_names: Optional[List[str]] = None) -> str:
        for i in range(len(table_names)):
            table_names[i] = extract_essential(table_names[i])
        return self._sql_agent.get_table_info_safe(table_names)

    def run_no_throw(self, command: str, fetch: str = "all") -> str:
        """Execute a SQL command and return the result as a string.

        This method catches any exceptions and returns an error message instead of raising an exception.

        Args:
            command: The SQL command to execute
            fetch: Whether to fetch 'all' results or just 'one'

        Returns:
            A string representation of the result or an error message
        """
        command = extract_essential(command)

        try:
            # Log the query for debugging
            logger.info(f"Executing SQL query: {command}")

            return self._sql_agent.query(command)

        except Exception as e:
            logger.error(f"Error executing SQL command: {str(e)}\n{traceback.format_exc()}")
            # If this is a knowledge base query, provide a more helpful error message
            if "knowledge_base" in command.lower() or any(
                kb in command for kb in self._sql_agent.get_usable_knowledge_base_names()
            ):
                return f"Error executing knowledge base query: {str(e)}. Please check that the knowledge base exists and your query syntax is correct."
            return f"Error: {str(e)}"

    def get_usable_knowledge_base_names(self) -> List[str]:
        """Get a list of usable knowledge base names.

        Returns:
            A list of knowledge base names that can be used in queries
        """
        try:
            return self._sql_agent.get_usable_knowledge_base_names()
        except Exception as e:
            logger.error(f"Error getting usable knowledge base names: {str(e)}")
            return []

    def check_knowledge_base_permission(self, name):
        """Get a list of usable knowledge base names.

        Returns:
            A list of knowledge base names that can be used in queries
        """

        return self._sql_agent.check_knowledge_base_permission(name)



import copy
from array import array
from typing import Any
from dataclasses import dataclass, field, MISSING

import numpy as np
import pandas as pd
from pandas.api import types as pd_types
import sqlalchemy.types as sqlalchemy_types

from mindsdb_sql_parser.ast import TableColumn

from mindsdb.utilities import log
from mindsdb.api.executor.exceptions import WrongArgumentError
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE


logger = log.getLogger(__name__)


def get_mysql_data_type_from_series(series: pd.Series, do_infer: bool = False) -> MYSQL_DATA_TYPE:
    """Maps pandas Series data type to corresponding MySQL data type.

    This function examines the dtype of a pandas Series and returns the appropriate
    MySQL data type enum value. For object dtypes, it can optionally attempt to infer
    a more specific type.

    Args:
        series (pd.Series): The pandas Series to determine the MySQL type for
        do_infer (bool): If True and series has object dtype, attempt to infer a more specific type

    Returns:
        MYSQL_DATA_TYPE: The corresponding MySQL data type enum value
    """
    dtype = series.dtype
    if pd_types.is_object_dtype(dtype) and do_infer is True:
        dtype = series.infer_objects().dtype

    if pd_types.is_object_dtype(dtype):
        return MYSQL_DATA_TYPE.TEXT
    if pd_types.is_datetime64_dtype(dtype):
        return MYSQL_DATA_TYPE.DATETIME
    if pd_types.is_string_dtype(dtype):
        return MYSQL_DATA_TYPE.TEXT
    if pd_types.is_bool_dtype(dtype):
        return MYSQL_DATA_TYPE.BOOL
    if pd_types.is_integer_dtype(dtype):
        return MYSQL_DATA_TYPE.INT
    if pd_types.is_numeric_dtype(dtype):
        return MYSQL_DATA_TYPE.FLOAT
    return MYSQL_DATA_TYPE.TEXT


def _dump_vector(value: Any) -> Any:
    if isinstance(value, array):
        return value.tolist()
    return value


@dataclass(kw_only=True, slots=True)
class Column:
    name: str = field(default=MISSING)
    alias: str | None = None
    table_name: str | None = None
    table_alias: str | None = None
    type: MYSQL_DATA_TYPE | None = None
    database: str | None = None
    flags: dict = None
    charset: str | None = None

    def __post_init__(self):
        if self.alias is None:
            self.alias = self.name
        if self.table_alias is None:
            self.table_alias = self.table_name

    def get_hash_name(self, prefix):
        table_name = self.table_name if self.table_alias is None else self.table_alias
        name = self.name if self.alias is None else self.alias

        name = f"{prefix}_{table_name}_{name}"
        return name


def rename_df_columns(df: pd.DataFrame, names: list | None = None) -> None:
    """Inplace rename of dataframe columns

    Args:
        df (pd.DataFrame): dataframe
        names (Optional[List]): columns names to set
    """
    if names is not None:
        df.columns = names
    else:
        df.columns = list(range(len(df.columns)))


class ResultSet:
    def __init__(
        self,
        columns: list[Column] | None = None,
        values: list[list] | None = None,
        df: pd.DataFrame | None = None,
        affected_rows: int | None = None,
        is_prediction: bool = False,
        mysql_types: list[MYSQL_DATA_TYPE] | None = None,
    ):
        """
        Args:
            columns: list of Columns
            values (List[List]): data of resultSet, have to be list of lists with length equal to column
            df (pd.DataFrame): injected dataframe, have to have enumerated columns and length equal to columns
            affected_rows (int): number of affected rows
        """
        if columns is None:
            columns = []
        self._columns = columns

        if df is None:
            if values is None:
                df = None
            else:
                df = pd.DataFrame(values)
        self._df = df

        self.affected_rows = affected_rows

        self.is_prediction = is_prediction

        self.mysql_types = mysql_types

    def __repr__(self):
        col_names = ", ".join([col.name for col in self._columns])

        return f"{self.__class__.__name__}({self.length()} rows, cols: {col_names})"

    def __len__(self) -> int:
        if self._df is None:
            return 0
        return len(self._df)

    def __getitem__(self, slice_val):
        # return resultSet with sliced dataframe
        df = self._df[slice_val]
        return ResultSet(columns=self.columns, df=df)

    # --- converters ---

    @classmethod
    def from_df(
        cls,
        df: pd.DataFrame,
        database=None,
        table_name=None,
        table_alias=None,
        is_prediction: bool = False,
        mysql_types: list[MYSQL_DATA_TYPE] | None = None,
    ):
        match mysql_types:
            case None:
                mysql_types = [None] * len(df.columns)
            case list() if len(mysql_types) != len(df.columns):
                raise WrongArgumentError(f"Mysql types length mismatch: {len(mysql_types)} != {len(df.columns)}")

        columns = [
            Column(name=column_name, table_name=table_name, table_alias=table_alias, database=database, type=mysql_type)
            for column_name, mysql_type in zip(df.columns, mysql_types)
        ]

        rename_df_columns(df)
        return cls(df=df, columns=columns, is_prediction=is_prediction, mysql_types=mysql_types)

    @classmethod
    def from_df_cols(cls, df: pd.DataFrame, columns_dict: dict[str, Column], strict: bool = True) -> "ResultSet":
        """Create ResultSet from dataframe and dictionary of columns

        Args:
            df (pd.DataFrame): dataframe
            columns_dict (dict[str, Column]): dictionary of columns
            strict (bool): if True, raise an error if a column is not found in columns_dict

        Returns:
            ResultSet: result set

        Raises:
            ValueError: if a column is not found in columns_dict and strict is True
        """
        alias_idx = {column.alias: column for column in columns_dict.values() if column.alias is not None}

        columns = []
        for column_name in df.columns:
            if strict and column_name not in columns_dict:
                raise ValueError(f"Column {column_name} not found in columns_dict")
            column = columns_dict.get(column_name) or alias_idx.get(column_name) or Column(name=column_name)
            columns.append(column)

        rename_df_columns(df)

        return cls(columns=columns, df=df)

    def to_df(self):
        columns_names = self.get_column_names()
        df = self.get_raw_df()
        rename_df_columns(df, columns_names)
        return df

    def to_df_cols(self, prefix: str = "") -> tuple[pd.DataFrame, dict[str, Column]]:
        # returns dataframe and dict of columns
        #   can be restored to ResultSet by from_df_cols method

        columns = []
        col_names = {}
        for col in self._columns:
            name = col.get_hash_name(prefix)
            columns.append(name)
            col_names[name] = col

        df = self.get_raw_df()
        rename_df_columns(df, columns)
        return df, col_names

    # --- tables ---

    def get_tables(self):
        tables_idx = []
        tables = []
        cols = ["database", "table_name", "table_alias"]
        for col in self._columns:
            table = (col.database, col.table_name, col.table_alias)
            if table not in tables_idx:
                tables_idx.append(table)
                tables.append(dict(zip(cols, table)))
        return tables

    # --- columns ---

    def get_col_index(self, col):
        """
        Get column index
        :param col: column object
        :return: index of column
        """

        col_idx = None
        for i, col0 in enumerate(self._columns):
            if col0 is col:
                col_idx = i
                break
        if col_idx is None:
            raise WrongArgumentError(f"Column is not found: {col}")
        return col_idx

    def add_column(self, col, values=None):
        self._columns.append(col)

        col_idx = len(self._columns) - 1
        if self._df is not None:
            self._df[col_idx] = values
        return col_idx

    def del_column(self, col):
        idx = self.get_col_index(col)
        self._columns.pop(idx)

        self._df.drop(idx, axis=1, inplace=True)
        rename_df_columns(self._df)

    @property
    def columns(self):
        return self._columns

    def get_column_names(self):
        columns = [col.name if col.alias is None else col.alias for col in self._columns]
        return columns

    def find_columns(self, alias=None, table_alias=None):
        col_list = []
        for col in self.columns:
            if alias is not None and col.alias.lower() != alias.lower():
                continue
            if table_alias is not None and col.table_alias.lower() != table_alias.lower():
                continue
            col_list.append(col)

        return col_list

    def copy_column_to(self, col, result_set2):
        # copy with values
        idx = self.get_col_index(col)

        values = [row[idx] for row in self._records]

        col2 = copy.deepcopy(col)

        result_set2.add_column(col2, values)
        return col2

    def set_col_type(self, col_idx, type_name):
        self.columns[col_idx].type = type_name
        if self._df is not None:
            self._df[col_idx] = self._df[col_idx].astype(type_name)

    # --- records ---

    def get_raw_df(self):
        if self._df is None:
            names = range(len(self._columns))
            return pd.DataFrame([], columns=names)
        return self._df

    def add_raw_df(self, df):
        if len(df.columns) != len(self._columns):
            raise WrongArgumentError(f"Record length mismatch columns length: {len(df.columns)} != {len(self.columns)}")

        rename_df_columns(df)

        if self._df is None:
            self._df = df
        else:
            self._df = pd.concat([self._df, df], ignore_index=True)

    def add_raw_values(self, values):
        # If some values are None, the DataFrame could have incorrect integer types, since 'NaN' is technically a float, so it will convert ints to floats automatically.
        df = pd.DataFrame(values).convert_dtypes(
            convert_integer=True,
            convert_floating=True,
            infer_objects=False,
            convert_string=False,
            convert_boolean=False,
        )
        self.add_raw_df(df)

    def get_ast_columns(self) -> list[TableColumn]:
        """Converts ResultSet columns to a list of TableColumn objects with SQLAlchemy types.

        This method processes each column in the ResultSet, determines its MySQL data type
        (inferring it if necessary), and maps it to the appropriate SQLAlchemy type.
        The resulting TableColumn objects most likely will be used in CREATE TABLE statement.

        Returns:
            list[TableColumn]: A list of TableColumn objects with properly mapped SQLAlchemy types
        """
        columns: list[TableColumn] = []

        type_mapping = {
            MYSQL_DATA_TYPE.TINYINT: sqlalchemy_types.INTEGER,
            MYSQL_DATA_TYPE.SMALLINT: sqlalchemy_types.INTEGER,
            MYSQL_DATA_TYPE.MEDIUMINT: sqlalchemy_types.INTEGER,
            MYSQL_DATA_TYPE.INT: sqlalchemy_types.INTEGER,
            MYSQL_DATA_TYPE.BIGINT: sqlalchemy_types.INTEGER,
            MYSQL_DATA_TYPE.YEAR: sqlalchemy_types.INTEGER,
            MYSQL_DATA_TYPE.BOOL: sqlalchemy_types.BOOLEAN,
            MYSQL_DATA_TYPE.BOOLEAN: sqlalchemy_types.BOOLEAN,
            MYSQL_DATA_TYPE.FLOAT: sqlalchemy_types.FLOAT,
            MYSQL_DATA_TYPE.DOUBLE: sqlalchemy_types.FLOAT,
            MYSQL_DATA_TYPE.TIME: sqlalchemy_types.Time,
            MYSQL_DATA_TYPE.DATE: sqlalchemy_types.Date,
            MYSQL_DATA_TYPE.DATETIME: sqlalchemy_types.DateTime,
            MYSQL_DATA_TYPE.TIMESTAMP: sqlalchemy_types.TIMESTAMP,
        }

        for i, column in enumerate(self._columns):
            column_type: MYSQL_DATA_TYPE | None = column.type

            # infer MYSQL_DATA_TYPE if not set
            if isinstance(column_type, MYSQL_DATA_TYPE) is False:
                if column_type is not None:
                    logger.warning(f"Unexpected column type: {column_type}")
                if self._df is None:
                    column_type = MYSQL_DATA_TYPE.TEXT
                else:
                    column_type = get_mysql_data_type_from_series(self._df.iloc[:, i])

            sqlalchemy_type = type_mapping.get(column_type, sqlalchemy_types.TEXT)

            columns.append(TableColumn(name=column.alias, type=sqlalchemy_type))
        return columns

    def to_lists(self, json_types=False):
        """
        :param type_cast: cast numpy types
            array->list, datetime64->str
        :return: list of lists
        """

        if len(self.get_raw_df()) == 0:
            return []
        # output for APIs. simplify types
        if json_types:
            df = self.get_raw_df().copy()
            for name, dtype in df.dtypes.to_dict().items():
                if pd.api.types.is_datetime64_any_dtype(dtype):
                    df[name] = df[name].dt.strftime("%Y-%m-%d %H:%M:%S.%f")
            for i, column in enumerate(self.columns):
                if column.type == MYSQL_DATA_TYPE.VECTOR:
                    df[i] = df[i].apply(_dump_vector)
            df.replace({np.nan: None}, inplace=True)
            return df.to_records(index=False).tolist()

        # slower but keep timestamp type
        df = self._df.replace({np.nan: None})  # TODO rework
        return df.to_dict("split")["data"]

    def get_column_values(self, col_idx):
        # get by column index
        df = self.get_raw_df()
        return list(df[df.columns[col_idx]])

    def set_column_values(self, col_name, values):
        # values is one value or list of values
        cols = self.find_columns(col_name)
        if len(cols) == 0:
            col_idx = self.add_column(Column(name=col_name))
        else:
            col_idx = self.get_col_index(cols[0])

        if self._df is not None:
            self._df[col_idx] = values

    def add_from_result_set(self, rs):
        source_names = rs.get_column_names()

        col_sequence = []
        for name in self.get_column_names():
            col_sequence.append(source_names.index(name))

        raw_df = rs.get_raw_df()[col_sequence]

        self.add_raw_df(raw_df)

    @property
    def records(self):
        return list(self.get_records())

    def get_records(self):
        # get records as dicts.
        # !!! Attention: !!!
        # if resultSet contents duplicate column name: only one of them will be in output
        names = self.get_column_names()
        for row in self.to_lists():
            yield dict(zip(names, row))

    def length(self):
        return len(self)



import dataclasses
import json
import os
import tempfile
import traceback
from datetime import datetime
from pathlib import Path

import lightwood
import pandas as pd
import requests
from lightwood.api.types import JsonAI
from pandas.core.frame import DataFrame

import mindsdb.utilities.profiler as profiler
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.integrations.utilities.utils import format_exception_error
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.fs import RESOURCE_GROUP, FileStorage
from mindsdb.interfaces.storage.json import get_json_storage
from mindsdb.utilities import log
from mindsdb.utilities.functions import mark_process

from .utils import brack_to_mod, rep_recur, unpack_jsonai_old_args

logger = log.getLogger(__name__)


def create_learn_mark():
    if os.name == 'posix':
        p = Path(tempfile.gettempdir()).joinpath('mindsdb/learn_processes/')
        p.mkdir(parents=True, exist_ok=True)
        p.joinpath(f'{os.getpid()}').touch()


def delete_learn_mark():
    if os.name == 'posix':
        p = (
            Path(tempfile.gettempdir())
            .joinpath('mindsdb/learn_processes/')
            .joinpath(f'{os.getpid()}')
        )
        if p.exists():
            p.unlink()


@mark_process(name='learn')
@profiler.profile()
def run_generate(df: DataFrame, predictor_id: int, model_storage, args: dict = None):

    model_storage.training_state_set(
        current_state_num=1, total_states=5, state_name='Generating problem definition'
    )
    json_ai_override = args.pop('using', {})

    if 'dtype_dict' in json_ai_override:
        args['dtype_dict'] = json_ai_override.pop('dtype_dict')

    if 'problem_definition' in json_ai_override:
        args = {**args, **json_ai_override['problem_definition']}

    if 'timeseries_settings' in args:
        for tss_key in [
            f.name for f in dataclasses.fields(lightwood.api.TimeseriesSettings)
        ]:
            k = f'timeseries_settings.{tss_key}'
            if k in json_ai_override:
                args['timeseries_settings'][tss_key] = json_ai_override.pop(k)

    problem_definition = lightwood.ProblemDefinition.from_dict(args)

    model_storage.training_state_set(
        current_state_num=2, total_states=5, state_name='Generating JsonAI'
    )
    json_ai = lightwood.json_ai_from_problem(df, problem_definition)
    json_ai = json_ai.to_dict()
    unpack_jsonai_old_args(json_ai_override)
    json_ai_override = brack_to_mod(json_ai_override)
    rep_recur(json_ai, json_ai_override)
    json_ai = JsonAI.from_dict(json_ai)

    model_storage.training_state_set(
        current_state_num=3, total_states=5, state_name='Generating code'
    )
    code = lightwood.code_from_json_ai(json_ai)

    predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
    predictor_record.code = code
    db.session.commit()

    json_storage = get_json_storage(resource_id=predictor_id)
    json_storage.set('json_ai', json_ai.to_dict())


@mark_process(name='learn')
@profiler.profile()
def run_fit(predictor_id: int, df: pd.DataFrame, model_storage) -> None:
    try:
        predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
        assert predictor_record is not None

        predictor_record.data = {'training_log': 'training'}
        predictor_record.status = PREDICTOR_STATUS.TRAINING
        db.session.commit()

        model_storage.training_state_set(
            current_state_num=4, total_states=5, state_name='Training model'
        )
        predictor: lightwood.PredictorInterface = lightwood.predictor_from_code(
            predictor_record.code
        )
        predictor.learn(df)

        db.session.refresh(predictor_record)

        fs = FileStorage(
            resource_group=RESOURCE_GROUP.PREDICTOR, resource_id=predictor_id, sync=True
        )
        predictor.save(fs.folder_path / fs.folder_name)
        fs.push(compression_level=0)

        predictor_record.data = predictor.model_analysis.to_dict()

        # getting training time for each tried model. it is possible to do
        # after training only
        fit_mixers = list(
            predictor.runtime_log[x]
            for x in predictor.runtime_log
            if isinstance(x, tuple) and x[0] == "fit_mixer"
        )
        submodel_data = predictor_record.data.get("submodel_data", [])
        # add training time to other mixers info
        if submodel_data and fit_mixers and len(submodel_data) == len(fit_mixers):
            for i, tr_time in enumerate(fit_mixers):
                submodel_data[i]["training_time"] = tr_time
        predictor_record.data["submodel_data"] = submodel_data

        model_storage.training_state_set(
            current_state_num=5, total_states=5, state_name='Complete'
        )
        predictor_record.dtype_dict = predictor.dtype_dict
        db.session.commit()
    except Exception as e:
        db.session.refresh(predictor_record)
        predictor_record.data = {'error': f'{traceback.format_exc()}\nMain error: {e}'}
        db.session.commit()
        raise e


@mark_process(name='learn')
def run_learn_remote(df: DataFrame, predictor_id: int) -> None:
    try:
        serialized_df = json.dumps(df.to_dict())
        predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
        resp = requests.post(
            predictor_record.data['train_url'],
            json={'df': serialized_df, 'target': predictor_record.to_predict[0]},
        )

        assert resp.status_code == 200
        predictor_record.data['status'] = 'complete'
    except Exception:
        predictor_record.data['status'] = 'error'
        predictor_record.data['error'] = str(resp.text)

    db.session.commit()


@mark_process(name='learn')
def run_learn(df: DataFrame, args: dict, model_storage) -> None:
    if df is None or df.shape[0] == 0:
        raise Exception(
            'No input data. Ensure the data source is healthy and try again.'
        )

    predictor_id = model_storage.predictor_id

    predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
    predictor_record.training_start_at = datetime.now()
    db.session.commit()

    run_generate(df, predictor_id, model_storage, args)
    run_fit(predictor_id, df, model_storage)

    predictor_record.status = PREDICTOR_STATUS.COMPLETE
    predictor_record.training_stop_at = datetime.now()
    db.session.commit()


@mark_process(name='finetune')
def run_finetune(df: DataFrame, args: dict, model_storage):
    try:
        if df is None or df.shape[0] == 0:
            raise Exception(
                'No input data. Ensure the data source is healthy and try again.'
            )

        base_predictor_id = args['base_model_id']
        base_predictor_record = db.Predictor.query.get(base_predictor_id)
        if base_predictor_record.status != PREDICTOR_STATUS.COMPLETE:
            raise Exception("Base model must be in status 'complete'")

        predictor_id = model_storage.predictor_id
        predictor_record = db.Predictor.query.get(predictor_id)

        # TODO move this to ModelStorage (don't work with database directly)
        predictor_record.data = {'training_log': 'training'}
        predictor_record.training_start_at = datetime.now()
        predictor_record.status = (
            PREDICTOR_STATUS.FINETUNING
        )  # TODO: parallel execution block
        db.session.commit()

        base_fs = FileStorage(
            resource_group=RESOURCE_GROUP.PREDICTOR,
            resource_id=base_predictor_id,
            sync=True,
        )
        predictor = lightwood.predictor_from_state(
            base_fs.folder_path / base_fs.folder_name, base_predictor_record.code
        )
        predictor.adjust(df, adjust_args=args.get('using', {}))

        fs = FileStorage(
            resource_group=RESOURCE_GROUP.PREDICTOR, resource_id=predictor_id, sync=True
        )
        predictor.save(fs.folder_path / fs.folder_name)
        fs.push(compression_level=0)

        predictor_record.data = (
            predictor.model_analysis.to_dict()
        )  # todo: update accuracy in LW as post-finetune hook
        predictor_record.code = base_predictor_record.code
        predictor_record.update_status = 'up_to_date'
        predictor_record.status = PREDICTOR_STATUS.COMPLETE
        predictor_record.training_stop_at = datetime.now()
        db.session.commit()

    except Exception as e:
        logger.error(e)
        predictor_id = model_storage.predictor_id
        predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
        logger.error(traceback.format_exc())
        error_message = format_exception_error(e)
        predictor_record.data = {"error": error_message}
        predictor_record.status = PREDICTOR_STATUS.ERROR
        db.session.commit()
        raise
    finally:
        if predictor_record.training_stop_at is None:
            predictor_record.training_stop_at = datetime.now()
            db.session.commit()



from typing import Any, Dict, List, Optional, Text

import pandas as pd
import salesforce_api
from salesforce_api.exceptions import AuthenticationError, RestRequestCouldNotBeUnderstoodError

from mindsdb.integrations.libs.api_handler import MetaAPIHandler
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    HandlerStatusResponse as StatusResponse,
    RESPONSE_TYPE,
)
from mindsdb.integrations.handlers.salesforce_handler.salesforce_tables import create_table_class
from mindsdb.integrations.handlers.salesforce_handler.constants import get_soql_instructions
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class SalesforceHandler(MetaAPIHandler):
    """
    This handler handles the connection and execution of SQL statements on Salesforce.
    """

    name = "salesforce"

    def __init__(self, name: Text, connection_data: Dict, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (Text): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the Salesforce API.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False
        self.thread_safe = True
        self.resource_names = []

    def connect(self) -> salesforce_api.client.Client:
        """
        Establishes a connection to the Salesforce API.

        Raises:
            ValueError: If the required connection parameters are not provided.
            AuthenticationError: If an authentication error occurs while connecting to the Salesforce API.

        Returns:
            salesforce_api.client.Client: A connection object to the Salesforce API.
        """
        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters.
        if not all(key in self.connection_data for key in ["username", "password", "client_id", "client_secret"]):
            raise ValueError("Required parameters (username, password, client_id, client_secret) must be provided.")

        try:
            self.connection = salesforce_api.Salesforce(
                username=self.connection_data["username"],
                password=self.connection_data["password"],
                client_id=self.connection_data["client_id"],
                client_secret=self.connection_data["client_secret"],
                is_sandbox=self.connection_data.get("is_sandbox", False),
            )
            self.is_connected = True

            resource_tables = self._get_resource_names()
            for resource_name in resource_tables:
                table_class = create_table_class(resource_name.lower())
                self._register_table(resource_name, table_class(self))

            return self.connection
        except AuthenticationError as auth_error:
            logger.error(f"Authentication error connecting to Salesforce, {auth_error}!")
            raise
        except Exception as unknown_error:
            logger.error(f"Unknwn error connecting to Salesforce, {unknown_error}!")
            raise

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Salesforce API.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except (AuthenticationError, ValueError) as known_error:
            logger.error(f"Connection check to Salesforce failed, {known_error}!")
            response.error_message = str(known_error)
        except Exception as unknown_error:
            logger.error(f"Connection check to Salesforce failed due to an unknown error, {unknown_error}!")
            response.error_message = str(unknown_error)

        self.is_connected = response.success

        return response

    def native_query(self, query: Text) -> Response:
        """
        Executes a native SOQL query on Salesforce and returns the result.

        Args:
            query (Text): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        connection = self.connect()

        try:
            results = connection.sobjects.query(query)

            parsed_results = []
            for result in results:
                del result["attributes"]

                # Check if the result contains any of the other Salesforce resources.
                if any(key in self.resource_names for key in result.keys()):
                    # Parse the result to extract the nested resources.
                    parsed_result = {}
                    for key, value in result.items():
                        if key in self.resource_names:
                            del value["attributes"]
                            parsed_result.update(
                                {f"{key}_{sub_key}": sub_value for sub_key, sub_value in value.items()}
                            )

                        else:
                            parsed_result[key] = value

                    parsed_results.append(parsed_result)

                else:
                    parsed_results.append(result)

            response = Response(RESPONSE_TYPE.TABLE, pd.DataFrame(parsed_results))
        except RestRequestCouldNotBeUnderstoodError as rest_error:
            logger.error(f"Error running query: {query} on Salesforce, {rest_error}!")
            response = Response(RESPONSE_TYPE.ERROR, error_code=0, error_message=str(rest_error))
        except Exception as unknown_error:
            logger.error(f"Error running query: {query} on Salesforce, {unknown_error}!")
            response = Response(RESPONSE_TYPE.ERROR, error_code=0, error_message=str(unknown_error))

        return response

    def _get_resource_names(self) -> List[str]:
        """
        Retrieves the names of the Salesforce resources with optimized pre-filtering.
        Returns:
            List[str]: A list of filtered resource names.
        """
        if not self.resource_names:
            # Check for user-specified table filtering first
            include_tables = self.connection_data.get("include_tables") or self.connection_data.get("tables")
            exclude_tables = self.connection_data.get("exclude_tables", [])

            if include_tables:
                # OPTIMIZATION: Skip expensive global describe() call
                # Only validate the specified tables
                logger.info(f"Using pre-filtered table list: {include_tables}")
                self.resource_names = self._validate_specified_tables(include_tables, exclude_tables)
            else:
                # Fallback to full discovery with hard-coded filtering
                logger.info("No table filter specified, performing full discovery...")
                self.resource_names = self._discover_all_tables_with_filtering(exclude_tables)

        return self.resource_names

    def _validate_specified_tables(self, include_tables: List[str], exclude_tables: List[str]) -> List[str]:
        """
        Validate user-specified tables without expensive global describe() call.

        Args:
            include_tables: List of table names to include
            exclude_tables: List of table names to exclude

        Returns:
            List[str]: Validated and filtered table names
        """
        validated_tables = []

        for table_name in include_tables:
            # Skip if explicitly excluded
            if table_name in exclude_tables:
                logger.info(f"Skipping excluded table: {table_name}")
                continue

            try:
                # Quick validation: check if table exists and is queryable
                # This is much faster than global describe()
                metadata = getattr(self.connection.sobjects, table_name).describe()
                if metadata.get("queryable", False):
                    validated_tables.append(table_name)
                    logger.debug(f"Validated table: {table_name}")
                else:
                    logger.warning(f"Table {table_name} is not queryable, skipping")
            except Exception as e:
                logger.warning(f"Table {table_name} not found or accessible: {e}")

        logger.info(f"Validated {len(validated_tables)} tables from include_tables")
        return validated_tables

    def _discover_all_tables_with_filtering(self, exclude_tables: List[str]) -> List[str]:
        """
        Fallback method: discover all tables with hard-coded filtering.

        Args:
            exclude_tables: List of table names to exclude

        Returns:
            List[str]: Filtered table names
        """
        # This is the original expensive approach - only used when no include_tables specified
        all_resources = [
            resource["name"]
            for resource in self.connection.sobjects.describe()["sobjects"]
            if resource.get("queryable", False)
        ]

        # Apply hard-coded filtering (existing logic)
        ignore_suffixes = ("Share", "History", "Feed", "ChangeEvent", "Tag", "Permission", "Setup", "Consent")
        ignore_prefixes = (
            "Apex",
            "CommPlatform",
            "Lightning",
            "Flow",
            "Transaction",
            "AI",
            "Aura",
            "ContentWorkspace",
            "Collaboration",
            "Datacloud",
        )
        ignore_exact = {
            "EntityDefinition",
            "FieldDefinition",
            "RecordType",
            "CaseStatus",
            "UserRole",
            "UserLicense",
            "UserPermissionAccess",
            "UserRecordAccess",
            "Folder",
            "Group",
            "Note",
            "ProcessDefinition",
            "ProcessInstance",
            "ContentFolder",
            "ContentDocumentSubscription",
            "DashboardComponent",
            "Report",
            "Dashboard",
            "Topic",
            "TopicAssignment",
            "Period",
            "Partner",
            "PackageLicense",
            "ColorDefinition",
            "DataUsePurpose",
            "DataUseLegalBasis",
        }

        ignore_substrings = (
            "CleanInfo",
            "Template",
            "Rule",
            "Definition",
            "Status",
            "Policy",
            "Setting",
            "Access",
            "Config",
            "Subscription",
            "DataType",
            "MilestoneType",
            "Entitlement",
            "Auth",
        )

        # Apply hard-coded filtering
        filtered = []
        for r in all_resources:
            if (
                not r.endswith(ignore_suffixes)
                and not r.startswith(ignore_prefixes)
                and not any(sub in r for sub in ignore_substrings)
                and r not in ignore_exact
                and r not in exclude_tables  # Apply user exclusions
            ):
                filtered.append(r)

        return filtered

    def meta_get_handler_info(self, **kwargs) -> str:
        """
        Retrieves information about the design and implementation of the API handler.
        This should include, but not be limited to, the following:
        - The type of SQL queries and operations that the handler supports.
        - etc.

        Args:
            kwargs: Additional keyword arguments that may be used in generating the handler information.

        Returns:
            str: A string containing information about the API handler's design and implementation.
        """
        return get_soql_instructions(self.name)

    def meta_get_tables(self, table_names: Optional[List[str]] = None) -> Response:
        """
        Retrieves metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (List): A list of table names for which to retrieve metadata.

        Returns:
            Response: A response object containing the table metadata.
        """
        connection = self.connect()

        # Retrieve the metadata for all Salesforce resources.
        main_metadata = connection.sobjects.describe()
        if table_names:
            # Filter the metadata for the specified tables.
            main_metadata = [
                resource for resource in main_metadata["sobjects"] if resource["name"].lower() in table_names
            ]
        else:
            main_metadata = main_metadata["sobjects"]

        return super().meta_get_tables(table_names=table_names, main_metadata=main_metadata)



"""
Constants for Salesforce handler.
"""


def get_soql_instructions(integration_name):
    return f"""This handler executes SOQL (Salesforce Object Query Language), NOT SQL! Follow these rules strictly:

**BASIC STRUCTURE:**
- NO "SELECT *" - must explicitly list all fields
  SQL: SELECT * FROM Account;
  SOQL: SELECT Id, Name, Industry FROM Account
- NO table aliases - use full table names only
  SQL: SELECT a.Name FROM Account a;
  SOQL: SELECT Name FROM Account
- NO column aliases - field names cannot be aliased
  SQL: SELECT Name AS CompanyName FROM Account;
  SOQL: SELECT Name FROM Account
- NO DISTINCT keyword - not supported in SOQL
  SQL: SELECT DISTINCT Industry FROM Account;
  SOQL: Not possible - use separate logic
- NO subqueries in FROM clause - only relationship-based subqueries allowed
  SQL: SELECT * FROM (SELECT Name FROM Account) AS AccountNames;
  SOQL: Not supported
- Do not use fields that are not defined in the schema or data catalog. Always reference exact field names.

**FIELD SELECTION:**
- Always include Id field when querying
  CORRECT: SELECT Id, Name, Industry FROM Account
  INCORRECT: SELECT Name, Industry FROM Account
- Field names are case-sensitive
  CORRECT: SELECT CreatedDate FROM Account
  INCORRECT: SELECT createddate FROM Account
- Use exact field names from the data catalog
  CORRECT: SELECT CustomerPriority__c FROM Account
  INCORRECT: SELECT customer_priority FROM Account

**FILTERING (WHERE clause):**
- Date/DateTime fields: Use unquoted literals in YYYY-MM-DD or YYYY-MM-DDThh:mm:ssZ format
  CORRECT: WHERE CloseDate >= 2025-05-28
  CORRECT: WHERE CreatedDate >= 2025-05-28T10:30:00Z
  INCORRECT: WHERE CloseDate >= '2025-05-28'
  INCORRECT: WHERE CreatedDate >= "2025-05-28"
- Special date literals: TODAY, YESTERDAY, LAST_WEEK, LAST_MONTH, LAST_QUARTER, LAST_YEAR, THIS_WEEK, THIS_MONTH, THIS_QUARTER, THIS_YEAR
  CORRECT: WHERE CreatedDate = TODAY
  CORRECT: WHERE LastModifiedDate >= LAST_MONTH
  CORRECT: WHERE CloseDate >= THIS_QUARTER
- Date arithmetic (e.g., TODAY - 10) is not supported. Use literals like LAST_N_DAYS:10 instead.
  CORRECT: WHERE CloseDate >= LAST_N_DAYS:10
  INCORRECT: WHERE CloseDate >= TODAY - 10
- LIKE operator: Only supports % wildcard, NO underscore (_) wildcard
  CORRECT: WHERE Name LIKE '%Corp%'
  CORRECT: WHERE Name LIKE 'Acme%'
  INCORRECT: WHERE Name LIKE 'A_me%'
- BETWEEN operator: NOT supported, use >= AND <= instead
  SQL: WHERE CreatedDate BETWEEN '2025-01-01' AND '2025-12-31'
  SOQL: WHERE CreatedDate >= 2025-01-01 AND CreatedDate <= 2025-12-31
- Boolean values: Use lowercase true/false, NOT TRUE/FALSE
  CORRECT: WHERE Active__c = true
  CORRECT: WHERE IsDeleted = false
  INCORRECT: WHERE Active__c = TRUE
  INCORRECT: WHERE IsDeleted = FALSE
- NULL values: Use lowercase null, NOT NULL
  CORRECT: WHERE ParentId = null
  CORRECT: WHERE Description != null
  INCORRECT: WHERE ParentId IS NULL
  INCORRECT: WHERE Description IS NOT NULL
- String values: Use single quotes for strings
  CORRECT: WHERE Industry = 'Technology'
  CORRECT: WHERE Name = 'Acme Corp'
  INCORRECT: WHERE Industry = "Technology"
- Multi-select picklist fields: Use INCLUDES('value1;value2') or EXCLUDES('value1;value2')
  CORRECT: WHERE Services__c INCLUDES ('Consulting;Support')
  CORRECT: WHERE Services__c EXCLUDES ('Training')
  INCORRECT: WHERE Services__c = 'Consulting'
- Limited subquery support - only IN/NOT IN with non-correlated subqueries in WHERE clause
  CORRECT: SELECT Id FROM Contact WHERE Id NOT IN (SELECT WhoId FROM Task)
  INCORRECT: SELECT Id FROM Contact WHERE NOT EXISTS (SELECT 1 FROM Task WHERE WhoId = Contact.Id)

**JOINS:**
- NO explicit JOIN syntax supported
  SQL: SELECT a.Name, c.FirstName FROM Account a JOIN Contact c ON a.Id = c.AccountId
  SOQL: Not supported - use relationship traversal (not applicable in this use case)

**AGGREGATES:**
- NO COUNT(*) - use COUNT(Id) instead
  SQL: SELECT COUNT(*) FROM Account
  SOQL: SELECT COUNT(Id) FROM Account
- Cannot mix aggregate functions with non-aggregate fields unless using GROUP BY
  CORRECT: SELECT Industry, COUNT(Id) FROM Account GROUP BY Industry
  CORRECT: SELECT COUNT(Id) FROM Account
  INCORRECT: SELECT Industry, Name, COUNT(Id) FROM Account
- NO GROUP_CONCAT or string aggregation functions
  SQL: SELECT GROUP_CONCAT(Name) FROM Account
  SOQL: Not supported
- NO HAVING clause
  SQL: SELECT Industry, COUNT(*) FROM Account GROUP BY Industry HAVING COUNT(*) > 5
  SOQL: Not supported - filter with separate logic
- GROUP BY has limited field type support
  CORRECT: SELECT Industry, COUNT(Id) FROM Account GROUP BY Industry
  INCORRECT: SELECT Description, COUNT(Id) FROM Account GROUP BY Description (textarea fields not supported)

**FUNCTIONS:**
- Date functions: CALENDAR_MONTH(), CALENDAR_YEAR(), CALENDAR_QUARTER(), DAY_IN_MONTH(), DAY_IN_WEEK(), DAY_IN_YEAR(), HOUR_IN_DAY(), WEEK_IN_MONTH(), WEEK_IN_YEAR()
  CORRECT: SELECT Id, Name FROM Account WHERE CALENDAR_YEAR(CreatedDate) = 2025
  CORRECT: SELECT Id, Name FROM Account WHERE CALENDAR_MONTH(CreatedDate) = 5
  CORRECT: SELECT Id, Name FROM Account WHERE DAY_IN_WEEK(CreatedDate) = 2
- NO math functions: ROUND, FLOOR, CEILING, ABS, etc.
  SQL: SELECT ROUND(AnnualRevenue, 2) FROM Account
  SOQL: Not supported
- NO conditional functions: CASE WHEN, COALESCE, NULLIF, etc.
  SQL: SELECT CASE WHEN Industry = 'Technology' THEN 'Tech' ELSE 'Other' END FROM Account
  SOQL: Not supported
- NO string functions except INCLUDES/EXCLUDES for multi-select picklists
  SQL: SELECT UPPER(Name) FROM Account
  SOQL: Not supported

**OPERATORS:**
- Supported: =, !=, <, >, <=, >=, LIKE, IN, NOT IN, INCLUDES, EXCLUDES
  CORRECT: WHERE Industry = 'Technology'
  CORRECT: WHERE AnnualRevenue >= 1000000
  CORRECT: WHERE Industry IN ('Technology', 'Finance')
  CORRECT: WHERE Industry NOT IN ('Government', 'Non-Profit')
  CORRECT: WHERE Services__c INCLUDES ('Consulting')
- NOT supported: REGEXP, BETWEEN, EXISTS, NOT EXISTS
  SQL: WHERE Name REGEXP '^[A-Z]'
  SOQL: Not supported

**SORTING & LIMITING:**
- ORDER BY: Fully supported
  CORRECT: SELECT Id, Name FROM Account ORDER BY Name ASC
  CORRECT: SELECT Id, Name FROM Account ORDER BY CreatedDate DESC, Name ASC
  CORRECT: SELECT Id, Name FROM Account ORDER BY Name NULLS LAST
- LIMIT: Maximum 2000 records, use smaller limits for better performance
  CORRECT: SELECT Id, Name FROM Account LIMIT 100
  CORRECT: SELECT Id, Name FROM Account LIMIT 2000
  INCORRECT: SELECT Id, Name FROM Account LIMIT 5000
- NO OFFSET: Not supported for pagination
  SQL: SELECT Id, Name FROM Account LIMIT 10 OFFSET 20
  SOQL: Not supported

**DATA TYPES:**
- picklist: Single-select dropdown, use = operator with string values
  CORRECT: WHERE Industry = 'Technology'
  CORRECT: WHERE Rating = 'Hot'
- reference: Foreign key field, typically ends with Id
  CORRECT: WHERE OwnerId = '00530000003OOwn'
  CORRECT: WHERE AccountId = '0013000000UzXyz'
- boolean: Use lowercase true/false
  CORRECT: WHERE IsDeleted = false
  CORRECT: WHERE Active__c = true
- currency: Numeric field for money values
  CORRECT: WHERE AnnualRevenue > 1000000
  CORRECT: WHERE AnnualRevenue >= 500000.50
- date: Date only, use YYYY-MM-DD format
  CORRECT: WHERE LastActivityDate = 2025-05-28
  CORRECT: WHERE SLAExpirationDate__c >= 2025-01-01
- datetime: Date and time, use YYYY-MM-DDThh:mm:ssZ format
  CORRECT: WHERE CreatedDate >= 2025-05-28T10:30:00Z
  CORRECT: WHERE LastModifiedDate = 2025-05-28T00:00:00Z
- double/int: Numeric fields
  CORRECT: WHERE NumberOfEmployees > 100
  CORRECT: WHERE NumberofLocations__c >= 5.5
- string/textarea: Text fields, use single quotes
  CORRECT: WHERE Name = 'Acme Corporation'
  CORRECT: WHERE Description = 'Leading tech company'
- phone/url/email: Specialized string fields, treat as strings
  CORRECT: WHERE Phone = '555-1234'
  CORRECT: WHERE Website = 'https://example.com'

**COMMON MISTAKES TO AVOID:**
- Using SELECT * (not allowed)
  WRONG: SELECT * FROM Account
  RIGHT: SELECT Id, Name, Industry FROM Account
- Quoting date literals (dates must be unquoted)
  WRONG: WHERE CreatedDate >= '2025-01-01'
  RIGHT: WHERE CreatedDate >= 2025-01-01
- Using SQL JOIN syntax (not supported)
  WRONG: SELECT Account.Name FROM Account JOIN Contact ON Account.Id = Contact.AccountId
  RIGHT: Use relationship traversal (not applicable in this use case)
- Using BETWEEN operator (not supported)
  WRONG: WHERE CreatedDate BETWEEN 2025-01-01 AND 2025-12-31
  RIGHT: WHERE CreatedDate >= 2025-01-01 AND CreatedDate <= 2025-12-31
- Using uppercase TRUE/FALSE/NULL (must be lowercase)
  WRONG: WHERE Active__c = TRUE
  RIGHT: WHERE Active__c = true
- Using underscore _ in LIKE patterns (only % supported)
  WRONG: WHERE Name LIKE 'A_me%'
  RIGHT: WHERE Name LIKE 'A%me%'
- Mixing aggregate and non-aggregate fields without GROUP BY
  WRONG: SELECT Name, COUNT(Id) FROM Account
  RIGHT: SELECT Industry, COUNT(Id) FROM Account GROUP BY Industry

**EXAMPLE QUERIES:**
- Basic selection: SELECT Id, Name, Industry FROM Account WHERE Industry = 'Technology'
- Date filtering: SELECT Id, Name FROM Account WHERE CreatedDate >= 2025-01-01
- Multiple conditions: SELECT Id, Name FROM Account WHERE Name LIKE '%Corp%' AND Industry IN ('Technology', 'Finance')
- Aggregation: SELECT Industry, COUNT(Id) FROM Account GROUP BY Industry
- Boolean and numeric: SELECT Id, Name FROM Account WHERE Active__c = true AND NumberOfEmployees > 100
- Date functions: SELECT Id, Name FROM Account WHERE CALENDAR_YEAR(CreatedDate) = 2025
- Null checks: SELECT Id, Name FROM Account WHERE ParentId = null
- Multi-select picklist: SELECT Id, Name FROM Account WHERE Services__c INCLUDES ('Consulting;Support')
- Sorting and limiting: SELECT Id, Name FROM Account ORDER BY Name ASC LIMIT 50


***EXECUTION INSTRUCTIONS. IMPORTANT!***
After generating the core SOQL (and nothing else), always make sure you wrap it exactly as:

    SELECT * 
      FROM {integration_name}(
        /* your generated SOQL goes here, without a trailing semicolon */
      )

Return only that wrapper call.
"""



from .__about__ import __version__ as version
from .__about__ import __description__ as description
from mindsdb.integrations.libs.const import HANDLER_TYPE
from mindsdb.utilities import log

logger = log.getLogger(__name__)


try:
    from .anthropic_handler import AnthropicHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "Anthropic"
name = "anthropic"
type = HANDLER_TYPE.ML
icon_path = 'icon.svg'
permanent = False
__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path"]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "Gong API key for authentication.",
        "secret": True,
        "required": False,
        "label": "API Key",
    },
    # Basic Authentication with Access Key + Secret Key (Option 2)
    access_key={
        "type": ARG_TYPE.STR,
        "description": "Gong Access Key for basic authentication (if not using OAuth).",
        "secret": True,
        "required": False,
        "label": "Access Key",
    },
    secret_key={
        "type": ARG_TYPE.PWD,
        "description": "Gong Secret Key for basic authentication (if not using OAuth).",
        "secret": True,
        "required": False,
        "label": "Secret Key",
    },
    base_url={
        "type": ARG_TYPE.STR,
        "description": "Gong API base URL (optional, defaults to production).",
        "required": False,
        "label": "Base URL",
    },
)

connection_args_example = OrderedDict(api_key="your_gong_api_key_here", base_url="https://api.gong.io")



from typing import Optional, Dict

import cohere
import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.utilities import log

from mindsdb.integrations.utilities.handler_utils import get_api_key


logger = log.getLogger(__name__)


class CohereHandler(BaseMLEngine):
    """
    Integration with the Cohere Python Library
    """
    name = 'cohere'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("Cohere engine requires a USING clause! Refer to its documentation for more details.")

        self.generative = True
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:

        args = self.model_storage.json_get('args')

        input_keys = list(args.keys())

        logger.info(f"Input keys: {input_keys}!")

        input_column = args['using']['column']

        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')

        result_df = pd.DataFrame()

        if args['using']['task'] == 'text-summarization':
            result_df['predictions'] = df[input_column].apply(self.predict_text_summary)

        elif args['using']['task'] == 'text-generation':
            result_df['predictions'] = df[input_column].apply(self.predict_text_generation)

        else:
            raise Exception(f"Task {args['using']['task']} is not supported!")

        result_df = result_df.rename(columns={'predictions': args['target']})

        return result_df

    def predict_text_summary(self, text):
        """
        connects with cohere api to predict the summary of the input text

        """

        args = self.model_storage.json_get('args')

        api_key = get_api_key('cohere', args["using"], self.engine_storage, strict=False)
        co = cohere.Client(api_key)

        response = co.summarize(text)
        text_summary = response.summary

        return text_summary

    def predict_text_generation(self, text):
        """
        connects with cohere api to predict the next prompt of the input text

        """
        args = self.model_storage.json_get('args')

        api_key = get_api_key('cohere', args["using"], self.engine_storage, strict=False)
        co = cohere.Client(api_key)

        response = co.generate(text)
        text_generated = response.generations[0].text

        return text_generated



from dataclasses import dataclass, astuple
import traceback
import json
import csv
from io import BytesIO, StringIO, IOBase
from pathlib import Path
import codecs
from typing import List, Generator

import filetype
import pandas as pd
from charset_normalizer import from_bytes
from mindsdb.interfaces.knowledge_base.preprocessing.text_splitter import TextSplitter

from mindsdb.utilities import log

logger = log.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 250


class FileProcessingError(Exception): ...


@dataclass(frozen=True, slots=True)
class _SINGLE_PAGE_FORMAT:
    CSV: str = "csv"
    JSON: str = "json"
    TXT: str = "txt"
    PDF: str = "pdf"
    PARQUET: str = "parquet"


SINGLE_PAGE_FORMAT = _SINGLE_PAGE_FORMAT()


@dataclass(frozen=True, slots=True)
class _MULTI_PAGE_FORMAT:
    XLSX: str = "xlsx"


MULTI_PAGE_FORMAT = _MULTI_PAGE_FORMAT()


def decode(file_obj: IOBase) -> StringIO:
    file_obj.seek(0)
    byte_str = file_obj.read()
    # Move it to StringIO
    try:
        # Handle Microsoft's BOM "special" UTF-8 encoding
        if byte_str.startswith(codecs.BOM_UTF8):
            data_str = StringIO(byte_str.decode("utf-8-sig"))
        else:
            file_encoding_meta = from_bytes(
                byte_str[: 32 * 1024],
                steps=32,  # Number of steps/block to extract from my_byte_str
                chunk_size=1024,  # Set block size of each extraction)
                explain=False,
            )
            best_meta = file_encoding_meta.best()
            errors = "strict"
            if best_meta is not None:
                encoding = file_encoding_meta.best().encoding

                try:
                    data_str = StringIO(byte_str.decode(encoding, errors))
                except UnicodeDecodeError:
                    encoding = "utf-8"
                    errors = "replace"

                    data_str = StringIO(byte_str.decode(encoding, errors))
            else:
                encoding = "utf-8"
                errors = "replace"

                data_str = StringIO(byte_str.decode(encoding, errors))
    except Exception as e:
        logger.error(traceback.format_exc())
        raise FileProcessingError("Could not load into string") from e

    return data_str


class FormatDetector:
    supported_formats = astuple(SINGLE_PAGE_FORMAT) + astuple(MULTI_PAGE_FORMAT)
    multipage_formats = astuple(MULTI_PAGE_FORMAT)

    def __init__(
        self,
        path: str | None = None,
        name: str | None = None,
        file: IOBase | None = None,
    ):
        """
        File format detector
        One of these arguments has to be passed: `path` or `file`

        :param path: path to the file
        :param name: name of the file
        :param file: file descriptor (via open(...), of BytesIO(...))
        """
        if path is not None:
            file = open(path, "rb")

        elif file is not None:
            if name is None:
                if hasattr(file, "name"):
                    path = file.name
                else:
                    path = "file"
        else:
            raise FileProcessingError("Wrong arguments: path or file is required")

        if name is None:
            name = Path(path).name

        self.name = name
        self.file_obj = file
        self.format = None

        self.parameters = {}

    def get_format(self) -> str:
        if self.format is not None:
            return self.format

        format = self.get_format_by_name()
        if format is not None:
            if format not in self.supported_formats:
                raise FileProcessingError(f"Not supported format: {format}")

        if format is None and self.file_obj is not None:
            format = self.get_format_by_content()
            self.file_obj.seek(0)

        if format is None:
            raise FileProcessingError(f"Unable to detect format: {self.name}")

        self.format = format
        return format

    def get_format_by_name(self):
        extension = Path(self.name).suffix.strip(".").lower()
        if extension == "tsv":
            extension = "csv"
            self.parameters["delimiter"] = "\t"

        return extension or None

    def get_format_by_content(self):
        if self.is_parquet(self.file_obj):
            return SINGLE_PAGE_FORMAT.PARQUET

        file_type = filetype.guess(self.file_obj)
        if file_type is not None:
            if file_type.mime in {
                "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
                "application/vnd.ms-excel",
            }:
                return MULTI_PAGE_FORMAT.XLSX

            if file_type.mime == "application/pdf":
                return SINGLE_PAGE_FORMAT.PDF

        file_obj = decode(self.file_obj)

        if self.is_json(file_obj):
            return SINGLE_PAGE_FORMAT.JSON

        if self.is_csv(file_obj):
            return SINGLE_PAGE_FORMAT.CSV

    @staticmethod
    def is_json(data_obj: StringIO) -> bool:
        # see if its JSON
        data_obj.seek(0)
        text = data_obj.read(100).strip()
        data_obj.seek(0)
        if len(text) > 0:
            # it looks like a json, then try to parse it
            if text.startswith("{") or text.startswith("["):
                try:
                    json.loads(data_obj.read())
                    return True
                except Exception:
                    return False
        return False

    @classmethod
    def is_csv(cls, data_obj: StringIO) -> bool:
        data_obj.seek(0)
        sample = data_obj.readline()  # trying to get dialect from header
        try:
            data_obj.seek(0)
            csv.Sniffer().sniff(sample)

            # Avoid a false-positive for json files
            if cls.is_json(data_obj):
                return False
            return True
        except Exception:
            return False

    @staticmethod
    def is_parquet(data: IOBase) -> bool:
        # Check first and last 4 bytes equal to PAR1.
        # Refer: https://parquet.apache.org/docs/file-format/
        parquet_sig = b"PAR1"
        data.seek(0, 0)
        start_meta = data.read(4)
        data.seek(-4, 2)
        end_meta = data.read()
        data.seek(0)
        if start_meta == parquet_sig and end_meta == parquet_sig:
            return True
        return False


def format_column_names(df: pd.DataFrame):
    df.columns = [column.strip(" \t") for column in df.columns]
    if len(df.columns) != len(set(df.columns)) or any(len(column_name) == 0 for column_name in df.columns):
        raise FileProcessingError("Each column should have a unique and non-empty name.")


class FileReader(FormatDetector):
    def _get_fnc(self):
        format = self.get_format()
        func = getattr(self, f"read_{format}", None)
        if func is None:
            raise FileProcessingError(f"Unsupported format: {format}")

        if format in astuple(MULTI_PAGE_FORMAT):

            def format_multipage(*args, **kwargs):
                for page_number, df in func(*args, **kwargs):
                    format_column_names(df)
                    yield page_number, df

            return format_multipage

        def format_singlepage(*args, **kwargs) -> pd.DataFrame:
            """Check that the columns have unique not-empty names"""
            df = func(*args, **kwargs)
            format_column_names(df)
            return df

        return format_singlepage

    def get_pages(self, **kwargs) -> List[str]:
        """
        Get list of tables in file
        """
        format = self.get_format()
        if format not in self.multipage_formats:
            # only one table
            return ["main"]

        func = self._get_fnc()
        self.file_obj.seek(0)

        return [name for name, _ in func(self.file_obj, only_names=True, **kwargs)]

    def get_contents(self, **kwargs) -> dict[str, pd.DataFrame]:
        """
        Get all info(pages with content) from file as dict: {tablename, content}
        """
        func = self._get_fnc()
        self.file_obj.seek(0)

        format = self.get_format()
        if format not in self.multipage_formats:
            # only one table
            return {"main": func(self.file_obj, name=self.name, **kwargs)}

        return {name: df for name, df in func(self.file_obj, **kwargs)}

    def get_page_content(self, page_name: str | None = None, **kwargs) -> pd.DataFrame:
        """
        Get content of a single table
        """
        func = self._get_fnc()
        self.file_obj.seek(0)

        format = self.get_format()
        if format not in self.multipage_formats:
            # only one table
            return func(self.file_obj, name=self.name, **kwargs)

        for _, df in func(self.file_obj, name=self.name, page_name=page_name, **kwargs):
            return df

    @staticmethod
    def _get_csv_dialect(buffer, delimiter: str | None = None) -> csv.Dialect | None:
        sample = buffer.readline()  # trying to get dialect from header
        buffer.seek(0)
        try:
            if isinstance(sample, bytes):
                sample = sample.decode()

            if delimiter is not None:
                accepted_csv_delimiters = [delimiter]
            else:
                accepted_csv_delimiters = [",", "\t", ";"]
            try:
                dialect = csv.Sniffer().sniff(sample, delimiters=accepted_csv_delimiters)
                dialect.doublequote = True  # assume that all csvs have " as string escape
            except Exception:
                dialect = csv.reader(sample).dialect
                if dialect.delimiter not in accepted_csv_delimiters:
                    raise FileProcessingError(f"CSV delimeter '{dialect.delimiter}' is not supported")

        except csv.Error:
            dialect = None
        return dialect

    @classmethod
    def read_csv(cls, file_obj: BytesIO, delimiter: str | None = None, **kwargs) -> pd.DataFrame:
        file_obj = decode(file_obj)
        dialect = cls._get_csv_dialect(file_obj, delimiter=delimiter)
        return pd.read_csv(file_obj, sep=dialect.delimiter, index_col=False)

    @staticmethod
    def read_txt(file_obj: BytesIO, name: str | None = None, **kwargs) -> pd.DataFrame:
        # the lib is heavy, so import it only when needed

        file_obj = decode(file_obj)

        text = file_obj.read()

        text_splitter = TextSplitter(chunk_size=DEFAULT_CHUNK_SIZE, chunk_overlap=DEFAULT_CHUNK_OVERLAP)

        docs = text_splitter.split_text(text)
        return pd.DataFrame([{"content": doc, "metadata": {"source_file": name, "file_format": "txt"}} for doc in docs])

    @staticmethod
    def read_pdf(file_obj: BytesIO, name: str | None = None, **kwargs) -> pd.DataFrame:
        # the libs are heavy, so import it only when needed
        import fitz  # pymupdf

        with fitz.open(stream=file_obj.read()) as pdf:  # open pdf
            text = chr(12).join([page.get_text() for page in pdf])

        text_splitter = TextSplitter(chunk_size=DEFAULT_CHUNK_SIZE, chunk_overlap=DEFAULT_CHUNK_OVERLAP)

        split_text = text_splitter.split_text(text)

        return pd.DataFrame(
            {
                "content": split_text,
                "metadata": [{"file_format": "pdf", "source_file": name}] * len(split_text),
            }
        )

    @staticmethod
    def read_json(file_obj: BytesIO, **kwargs) -> pd.DataFrame:
        file_obj = decode(file_obj)
        file_obj.seek(0)
        json_doc = json.loads(file_obj.read())
        return pd.json_normalize(json_doc, max_level=0)

    @staticmethod
    def read_parquet(file_obj: BytesIO, **kwargs) -> pd.DataFrame:
        return pd.read_parquet(file_obj)

    @staticmethod
    def read_xlsx(
        file_obj: BytesIO,
        page_name: str | None = None,
        only_names: bool = False,
        **kwargs,
    ) -> Generator[tuple[str, pd.DataFrame | None], None, None]:
        with pd.ExcelFile(file_obj) as xls:
            if page_name is not None:
                # return specific page
                yield page_name, pd.read_excel(xls, sheet_name=page_name)

            for page_name in xls.sheet_names:
                if only_names:
                    # extract only pages names
                    df = None
                else:
                    df = pd.read_excel(xls, sheet_name=page_name)
                yield page_name, df



import imaplib
import email
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from datetime import datetime, timedelta

import pandas as pd
from mindsdb.integrations.handlers.email_handler.settings import EmailSearchOptions, EmailConnectionDetails
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class EmailClient:
    '''Class for searching emails using IMAP (Internet Messaging Access Protocol)'''

    _DEFAULT_SINCE_DAYS = 10

    def __init__(
            self,
            connection_data: EmailConnectionDetails
    ):
        self.email = connection_data.email
        self.password = connection_data.password
        self.imap_server = imaplib.IMAP4_SSL(connection_data.imap_server)
        self.smtp_server = smtplib.SMTP(connection_data.smtp_server, connection_data.smtp_port)

    def select_mailbox(self, mailbox: str = 'INBOX'):
        '''Logs in & selects a mailbox from IMAP server. Defaults to INBOX, which is the default inbox.

        Parameters:
            mailbox (str): The name of the mailbox to select.
        '''
        ok, resp = self.imap_server.login(self.email, self.password)
        if ok != 'OK':
            raise ValueError(
                f'Unable to login to mailbox {mailbox}. Please check your credentials: {str(resp)}')

        logger.info(f'Logged in to mailbox {mailbox}')

        ok, resp = self.imap_server.select(mailbox)
        if ok != 'OK':
            raise ValueError(
                f'Unable to select mailbox {mailbox}. Please check the mailbox name: {str(resp)}')

        logger.info(f'Selected mailbox {mailbox}')

    def logout(self):
        '''Shuts down the connection to the IMAP and SMTP server.'''

        try:
            ok, resp = self.imap_server.logout()
            if ok != 'BYE':
                logger.error(
                    f'Unable to logout of IMAP client: {str(resp)}')
            logger.info('Logged out of IMAP server')
        except Exception as e:
            logger.error(
                f'Exception occurred while logging out from IMAP server: {str(e)}')

        try:
            self.smtp_server.quit()
            logger.info('Logged out of SMTP server')
        except Exception as e:
            logger.error(
                f'Exception occurred while logging out from SMTP server: {str(e)}')

    def send_email(self, to_addr: str, subject: str, body: str):
        '''
        Sends an email to the given address.

        Parameters:
            to_addr (str): The email address to send the email to.
            subject (str): The subject of the email.
            body (str): The body of the email.
        '''

        msg = MIMEMultipart()
        msg['From'] = self.email
        msg['To'] = to_addr
        msg['Subject'] = subject
        msg.attach(MIMEText(body, 'plain'))

        self.smtp_server.starttls()
        self.smtp_server.login(self.email, self.password)
        self.smtp_server.send_message(msg)
        logger.info(f'Email sent to {to_addr} with subject: {subject}')

    def search_email(self, options: EmailSearchOptions) -> pd.DataFrame:
        '''Searches emails based on the given options and returns a DataFrame.

        Parameters:
            options (EmailSearchOptions): Options to use when searching using IMAP.

        Returns:
            df (pd.DataFrame): A dataframe of emails resulting from the search.
        '''
        self.select_mailbox(options.mailbox)

        try:

            query_parts = []
            if options.subject is not None:
                query_parts.append(f'(SUBJECT "{options.subject}")')

            if options.to_field is not None:
                query_parts.append(f'(TO "{options.to_field}")')

            if options.from_field is not None:
                query_parts.append(f'(FROM "{options.from_field}")')

            if options.since_date is not None:
                since_date_str = options.since_date.strftime('%d-%b-%Y')
            else:
                since_date = datetime.today() - timedelta(days=EmailClient._DEFAULT_SINCE_DAYS)
                since_date_str = since_date.strftime('%d-%b-%Y')
            query_parts.append(f'(SINCE "{since_date_str}")')

            if options.until_date is not None:
                until_date_str = options.until_date.strftime('%d-%b-%Y')
                query_parts.append(f'(BEFORE "{until_date_str}")')

            if options.since_email_id is not None:
                query_parts.append(f'(UID {options.since_email_id}:*)')

            query = ' '.join(query_parts)
            ret = []
            _, items = self.imap_server.uid('search', None, query)
            items = items[0].split()
            for emailid in items:
                _, data = self.imap_server.uid('fetch', emailid, '(RFC822)')
                email_message = email.message_from_bytes(data[0][1])

                email_line = {}
                email_line['id'] = emailid.decode()
                email_line['to_field'] = email_message.get('To')
                email_line['from_field'] = email_message.get('From')
                email_line['subject'] = email_message.get('Subject')
                email_line['date'] = email_message.get('Date')

                plain_payload = None
                html_payload = None
                content_type = 'html'
                for part in email_message.walk():
                    subtype = part.get_content_subtype()
                    if subtype == 'plain':
                        # Prioritize plain text payloads when present.
                        plain_payload = part.get_payload(decode=True)
                        content_type = 'plain'
                        break
                    if subtype == 'html':
                        html_payload = part.get_payload(decode=True)
                body = plain_payload or html_payload
                if body is None:
                    # Very rarely messages won't have plain text or html payloads.
                    continue
                email_line['body'] = plain_payload or html_payload
                email_line['body_content_type'] = content_type
                ret.append(email_line)
        except Exception as e:
            raise Exception('Error searching email') from e

        return pd.DataFrame(ret)



import json
import logging
from logging.config import dictConfig

from mindsdb.utilities.config import config as app_config


logging_initialized = False


class JsonFormatter(logging.Formatter):
    def format(self, record):
        record_message = super().format(record)
        log_record = {
            "process_name": record.processName,
            "name": record.name,
            "message": record_message,
            "level": record.levelname,
            "time": record.created,
        }
        return json.dumps(log_record)


class ColorFormatter(logging.Formatter):
    green = "\x1b[32;20m"
    default = "\x1b[39;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s %(processName)15s %(levelname)-8s %(name)s: %(message)s"

    FORMATS = {
        logging.DEBUG: logging.Formatter(green + format + reset),
        logging.INFO: logging.Formatter(default + format + reset),
        logging.WARNING: logging.Formatter(yellow + format + reset),
        logging.ERROR: logging.Formatter(red + format + reset),
        logging.CRITICAL: logging.Formatter(bold_red + format + reset),
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        return log_fmt.format(record)


FORMATTERS = {
    "default": {"()": ColorFormatter},
    "json": {"()": JsonFormatter},
    "file": {"format": "%(asctime)s %(processName)15s %(levelname)-8s %(name)s: %(message)s"},
}


def get_console_handler_config_level() -> int:
    console_handler_config = app_config["logging"]["handlers"]["console"]
    return getattr(logging, console_handler_config["level"])


def get_file_handler_config_level() -> int:
    file_handler_config = app_config["logging"]["handlers"]["file"]
    return getattr(logging, file_handler_config["level"])


def get_mindsdb_log_level() -> int:
    console_handler_config_level = get_console_handler_config_level()
    file_handler_config_level = get_file_handler_config_level()

    return min(console_handler_config_level, file_handler_config_level)


def get_handlers_config(process_name: str) -> dict:
    handlers_config = {}
    console_handler_config = app_config["logging"]["handlers"]["console"]
    console_handler_config_level = getattr(logging, console_handler_config["level"])
    if console_handler_config["enabled"] is True:
        handlers_config["console"] = {
            "class": "logging.StreamHandler",
            "formatter": console_handler_config.get("formatter", "default"),
            "level": console_handler_config_level,
        }

    file_handler_config = app_config["logging"]["handlers"]["file"]
    file_handler_config_level = getattr(logging, file_handler_config["level"])
    if file_handler_config["enabled"] is True:
        file_name = file_handler_config["filename"]
        if process_name is not None:
            if "." in file_name:
                parts = file_name.rpartition(".")
                file_name = f"{parts[0]}_{process_name}.{parts[2]}"
            else:
                file_name = f"{file_name}_{process_name}"
        handlers_config["file"] = {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "file",
            "level": file_handler_config_level,
            "filename": app_config.paths["log"] / file_name,
            "maxBytes": file_handler_config["maxBytes"],  # 0.5 Mb
            "backupCount": file_handler_config["backupCount"],
        }
    return handlers_config


def configure_logging(process_name: str = None):
    handlers_config = get_handlers_config(process_name)
    mindsdb_log_level = get_mindsdb_log_level()

    logging_config = dict(
        version=1,
        formatters=FORMATTERS,
        handlers=handlers_config,
        loggers={
            "": {  # root logger
                "handlers": list(handlers_config.keys()),
                "level": mindsdb_log_level,
            },
            "__main__": {
                "level": mindsdb_log_level,
            },
            "mindsdb": {
                "level": mindsdb_log_level,
            },
            "alembic": {
                "level": mindsdb_log_level,
            },
        },
    )

    dictConfig(logging_config)


def initialize_logging(process_name: str = None) -> None:
    """Initialyze logging"""
    global logging_initialized
    if not logging_initialized:
        configure_logging(process_name)
        logging_initialized = True


# I would prefer to leave code to use logging.getLogger(), but there are a lot of complicated situations
# in MindsDB with processes being spawned that require logging to be configured again in a lot of cases.
# Using a custom logger-getter like this lets us do that logic here, once.
def getLogger(name=None):
    """
    Get a new logger, configuring logging first if it hasn't been done yet.
    """
    initialize_logging()
    return logging.getLogger(name)



import re
import html
import asyncio
from typing import List, Dict, Optional, Any

import pandas as pd
from mindsdb.interfaces.knowledge_base.preprocessing.text_splitter import TextSplitter

from mindsdb.integrations.utilities.rag.splitters.file_splitter import (
    FileSplitter,
    FileSplitterConfig,
)
from mindsdb.interfaces.agents.langchain_agent import create_chat_model
from mindsdb.interfaces.knowledge_base.preprocessing.models import (
    PreprocessingConfig,
    ProcessedChunk,
    PreprocessorType,
    ContextualConfig,
    Document,
    TextChunkingConfig,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)

_DEFAULT_CONTENT_COLUMN_NAME = "content"


class DocumentPreprocessor:
    """Base class for document preprocessing"""

    def __init__(self):
        """Initialize preprocessor"""
        self.splitter = None  # Will be set by child classes
        self.config = None

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        """Base implementation - should be overridden by child classes

        Args:
            documents: List of documents to process
        """
        raise NotImplementedError("Subclasses must implement process_documents")

    def _split_document(self, doc: Document) -> List[Document]:
        """Split document into chunks while preserving metadata"""
        if self.splitter is None:
            raise ValueError("Splitter not configured")

        metadata = doc.metadata or {}
        # Split and convert back to our Document type
        split_texts = self.splitter.split_text(doc.content)
        return [Document(content=text, metadata=metadata) for text in split_texts]

    def _get_source(self) -> str:
        """Get the source identifier for this preprocessor"""
        return self.__class__.__name__

    def to_dataframe(self, chunks: List[ProcessedChunk]) -> pd.DataFrame:
        """Convert processed chunks to dataframe format"""
        return pd.DataFrame([chunk.model_dump() for chunk in chunks])

    @staticmethod
    def dict_to_document(data: Dict[str, Any]) -> Document:
        """Convert a dictionary to a Document object"""
        return Document(
            id=data.get("id"),
            content=data.get("content", ""),
            embeddings=data.get("embeddings"),
            metadata=data.get("metadata", {}),
        )

    def _generate_chunk_id(
        self,
        chunk_index: Optional[int] = None,
        total_chunks: Optional[int] = None,
        start_char: Optional[int] = None,
        end_char: Optional[int] = None,
        provided_id: str = None,
        content_column: str = None,
    ) -> str:
        """Generate human-readable deterministic ID for a chunk
        Format: <doc_id>:<content_column>:<chunk_number>of<total_chunks>:<start_char>to<end_char>
        """
        if provided_id is None:
            raise ValueError("Document ID must be provided for chunk ID generation")

        if content_column is None:
            raise ValueError("Content column must be provided for chunk ID generation")

        chunk_id = f"{provided_id}:{content_column}:{chunk_index + 1}of{total_chunks}:{start_char}to{end_char}"
        logger.debug(f"Generated chunk ID: {chunk_id}")
        return chunk_id

    def _prepare_chunk_metadata(
        self,
        doc_id: Optional[str],
        chunk_index: Optional[int],
        base_metadata: Optional[Dict] = None,
    ) -> Dict:
        """Centralized method for preparing chunk metadata"""
        metadata = base_metadata or {}

        # Always preserve original document ID
        if doc_id is not None:
            metadata[self.config.doc_id_column_name] = doc_id

        # Add chunk index only for multi-chunk cases
        if chunk_index is not None:
            metadata["_chunk_index"] = chunk_index

        # Always set source
        metadata["_source"] = self._get_source()

        return metadata


class ContextualPreprocessor(DocumentPreprocessor):
    """Contextual preprocessing implementation that enhances document chunks with context"""

    DEFAULT_CONTEXT_TEMPLATE = """
<document>
{WHOLE_DOCUMENT}
</document>
Here is the chunk we want to situate within the whole document
<chunk>
{CHUNK_CONTENT}
</chunk>
Please give a short succinct context to situate this chunk within the overall document for the purposes of improving search retrieval of the chunk. Answer only with the succinct context and nothing else."""

    def __init__(self, config: ContextualConfig):
        """Initialize with contextual configuration"""
        super().__init__()
        self.config = config
        self.splitter = FileSplitter(
            FileSplitterConfig(chunk_size=config.chunk_size, chunk_overlap=config.chunk_overlap)
        )
        self.llm = create_chat_model(
            {
                "model_name": self.config.llm_config.model_name,
                "provider": self.config.llm_config.provider,
                **self.config.llm_config.params,
            }
        )
        self.context_template = config.context_template or self.DEFAULT_CONTEXT_TEMPLATE
        self.summarize = self.config.summarize

    def _prepare_prompts(self, chunk_contents: list[str], full_documents: list[str]) -> list[str]:
        def tag_replacer(match):
            tag = match.group(0)
            if tag.lower() not in ["<document>", "</document>", "<chunk>", "</chunk>"]:
                return tag
            return html.escape(tag)

        tag_pattern = r"</?document>|</?chunk>"
        prompts = []
        for chunk_content, full_document in zip(chunk_contents, full_documents):
            chunk_content = re.sub(tag_pattern, tag_replacer, chunk_content, flags=re.IGNORECASE)
            full_document = re.sub(tag_pattern, tag_replacer, full_document, flags=re.IGNORECASE)
            prompts.append(
                self.DEFAULT_CONTEXT_TEMPLATE.format(WHOLE_DOCUMENT=full_document, CHUNK_CONTENT=chunk_content)
            )

        return prompts

    def _generate_context(self, chunk_contents: list[str], full_documents: list[str]) -> list[str]:
        """Generate contextual description for a chunk using LLM"""
        prompts = self._prepare_prompts(chunk_contents, full_documents)

        # Check if LLM supports async
        if hasattr(self.llm, "abatch"):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                response = loop.run_until_complete(self._async_generate(prompts))
            finally:
                loop.close()
        else:
            # Use sync batch for non-async LLMs
            response = [resp.content for resp in self.llm.batch(prompts)]
        return response

    async def _async_generate(self, prompts: list[str]) -> list[str]:
        """Helper for async LLM generation"""
        return [resp.content for resp in await self.llm.abatch(prompts)]

    def _split_document(self, doc: Document) -> List[Document]:
        """Split document into chunks while preserving metadata"""
        # Use base class implementation
        return super()._split_document(doc)

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        chunks_list = []
        doc_index_list = []
        chunk_index_list = []
        processed_chunks = []

        for doc_index, doc in enumerate(documents):
            # Document ID must be provided by this point
            if doc.id is None:
                raise ValueError("Document ID must be provided before preprocessing")

            # Skip empty or whitespace-only content
            if not doc.content or not doc.content.strip():
                continue

            chunk_docs = self._split_document(doc)

            # Single chunk case
            if len(chunk_docs) == 1:
                chunk_doc = chunk_docs[0]
                if not chunk_doc.content or not chunk_doc.content.strip():
                    continue
                else:
                    chunks_list.append(chunk_doc)
                    doc_index_list.append(doc_index)
                    chunk_index_list.append(0)

            else:
                # Multiple chunks case
                for i, chunk_doc in enumerate(chunk_docs):
                    if not chunk_doc.content or not chunk_doc.content.strip():
                        continue
                    else:
                        chunks_list.append(chunk_doc)
                        doc_index_list.append(doc_index)
                        chunk_index_list.append(i)

        # Generate contexts
        doc_contents = [documents[i].content for i in doc_index_list]
        chunk_contents = [chunk_doc.content for chunk_doc in chunks_list]
        contexts = self._generate_context(chunk_contents, doc_contents)

        for context, chunk_doc, chunk_index, doc_index in zip(contexts, chunks_list, chunk_index_list, doc_index_list):
            processed_content = context if self.summarize else f"{context}\n\n{chunk_doc.content}"
            doc = documents[doc_index]

            # Initialize metadata
            metadata = {}
            if doc.metadata:
                metadata.update(doc.metadata)

            # Get content_column from metadata or use default
            content_column = metadata.get("_content_column")
            if content_column is None:
                # If content_column is not in metadata, use the default column name
                content_column = _DEFAULT_CONTENT_COLUMN_NAME
                logger.debug(f"No content_column found in metadata, using default: {_DEFAULT_CONTENT_COLUMN_NAME}")

            chunk_id = self._generate_chunk_id(
                chunk_index=chunk_index, provided_id=doc.id, content_column=content_column
            )
            processed_chunks.append(
                ProcessedChunk(
                    id=chunk_id,
                    content=processed_content,  # Use the content with context
                    embeddings=doc.embeddings,
                    metadata=self._prepare_chunk_metadata(doc.id, None, metadata),
                )
            )

        return processed_chunks


class TextChunkingPreprocessor(DocumentPreprocessor):
    """Default text chunking preprocessor using TextSplitter"""

    def __init__(self, config: Optional[TextChunkingConfig] = None):
        """Initialize with text chunking configuration"""
        super().__init__()
        self.config = config or TextChunkingConfig()
        self.splitter = TextSplitter(
            chunk_size=self.config.chunk_size,
            chunk_overlap=self.config.chunk_overlap,
            separators=self.config.separators,
        )

    def _split_document(self, doc: Document) -> List[Document]:
        """Split document into chunks while preserving metadata"""
        # Use base class implementation
        return super()._split_document(doc)

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        processed_chunks = []

        for doc in documents:
            # Document ID must be provided by this point
            if doc.id is None:
                raise ValueError("Document ID must be provided before preprocessing")

            # Skip empty or whitespace-only content
            if not doc.content or not doc.content.strip():
                continue

            chunk_docs = self._split_document(doc)
            total_chunks = len(chunk_docs)

            # Track character positions
            current_pos = 0
            for i, chunk_doc in enumerate(chunk_docs):
                if not chunk_doc.content or not chunk_doc.content.strip():
                    continue

                # Calculate chunk positions
                start_char = current_pos
                end_char = start_char + len(chunk_doc.content)
                current_pos = end_char + 1  # +1 for separator

                # Initialize metadata
                metadata = {}
                if doc.metadata:
                    metadata.update(doc.metadata)

                # Add position metadata
                metadata["_start_char"] = start_char
                metadata["_end_char"] = end_char

                # Get content_column from metadata or use default
                content_column = None
                if doc.metadata:
                    content_column = doc.metadata.get("_content_column")

                if content_column is None:
                    # If content_column is not in metadata, use the default column name
                    content_column = _DEFAULT_CONTENT_COLUMN_NAME
                    logger.debug(f"No content_column found in metadata, using default: {_DEFAULT_CONTENT_COLUMN_NAME}")

                chunk_id = self._generate_chunk_id(
                    chunk_index=i,
                    total_chunks=total_chunks,
                    start_char=start_char,
                    end_char=end_char,
                    provided_id=doc.id,
                    content_column=content_column,
                )

                processed_chunks.append(
                    ProcessedChunk(
                        id=chunk_id,
                        content=chunk_doc.content,
                        embeddings=doc.embeddings,
                        metadata=self._prepare_chunk_metadata(doc.id, i, metadata),
                    )
                )

        return processed_chunks


class PreprocessorFactory:
    """Factory for creating preprocessors based on configuration"""

    @staticmethod
    def create_preprocessor(
        config: Optional[PreprocessingConfig] = None,
    ) -> DocumentPreprocessor:
        """
        Create appropriate preprocessor based on configuration
        :param config: Preprocessing configuration
        :return: Configured preprocessor instance
        :raises ValueError: If unknown preprocessor type specified
        """
        if config is None:
            # Default to text chunking if no config provided
            return TextChunkingPreprocessor()

        if config.type == PreprocessorType.TEXT_CHUNKING:
            return TextChunkingPreprocessor(config.text_chunking_config)
        elif config.type == PreprocessorType.CONTEXTUAL:
            return ContextualPreprocessor(config.contextual_config)
        elif config.type == PreprocessorType.JSON_CHUNKING:
            # Import here to avoid circular imports
            from mindsdb.interfaces.knowledge_base.preprocessing.json_chunker import JSONChunkingPreprocessor

            return JSONChunkingPreprocessor(config.json_chunking_config)
        else:
            raise ValueError(f"Unknown preprocessor type: {config.type}")



from typing import Any, List, Union, Optional, Dict

from langchain_community.vectorstores import PGVector
from langchain_community.vectorstores.pgvector import Base

from pgvector.sqlalchemy import SPARSEVEC, Vector
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSON

from sqlalchemy.orm import Session


_generated_sa_tables = {}


class PGVectorMDB(PGVector):
    """
    langchain_community.vectorstores.PGVector adapted for mindsdb vector store table structure
    """

    def __init__(self, *args, is_sparse: bool = False, vector_size: Optional[int] = None, **kwargs):
        # todo get is_sparse and vector_size from kb vector table
        self.is_sparse = is_sparse
        if is_sparse and vector_size is None:
            raise ValueError("vector_size is required when is_sparse=True")
        self.vector_size = vector_size
        super().__init__(*args, **kwargs)

    def __post_init__(
        self,
    ) -> None:

        collection_name = self.collection_name

        if collection_name not in _generated_sa_tables:

            class EmbeddingStore(Base):
                """Embedding store."""

                __tablename__ = collection_name

                id = sa.Column(sa.Integer, primary_key=True)
                embedding = sa.Column(
                    "embeddings",
                    SPARSEVEC() if self.is_sparse else Vector() if self.vector_size is None else
                    SPARSEVEC(self.vector_size) if self.is_sparse else Vector(self.vector_size)
                )
                document = sa.Column("content", sa.String, nullable=True)
                cmetadata = sa.Column("metadata", JSON, nullable=True)

            _generated_sa_tables[collection_name] = EmbeddingStore

        self.EmbeddingStore = _generated_sa_tables[collection_name]

    def __query_collection(
            self,
            embedding: Union[List[float], Dict[int, float], str],
            k: int = 4,
            filter: Optional[Dict[str, str]] = None,
    ) -> List[Any]:
        """Query the collection."""
        with Session(self._bind) as session:
            if self.is_sparse:
                # Sparse vectors: expect string in format "{key:value,...}/size" or dictionary
                if isinstance(embedding, dict):
                    from pgvector.utils import SparseVector
                    embedding = SparseVector(embedding, self.vector_size)
                    embedding_str = embedding.to_text()
                elif isinstance(embedding, str):
                    # Use string as is - it should already be in the correct format
                    embedding_str = embedding
                # Use inner product for sparse vectors
                distance_op = "<#>"
                # For inner product, larger values are better matches
                order_direction = "ASC"
            else:
                # Dense vectors: expect string in JSON array format or list of floats
                if isinstance(embedding, list):
                    embedding_str = f"[{','.join(str(x) for x in embedding)}]"
                elif isinstance(embedding, str):
                    embedding_str = embedding
                # Use cosine similarity for dense vectors
                distance_op = "<=>"
                # For cosine similarity, smaller values are better matches
                order_direction = "ASC"

            # Use SQL directly for vector comparison
            query = sa.text(
                f"""
            SELECT t.*, t.embeddings {distance_op} '{embedding_str}' as distance
            FROM {self.collection_name} t
            ORDER BY distance {order_direction}
            LIMIT {k}
            """
            )
            results = session.execute(query).all()

            # Convert results to the expected format
            formatted_results = []
            for rec in results:
                metadata = rec.metadata if bool(rec.metadata) else {0: 0}
                embedding_store = self.EmbeddingStore()
                embedding_store.document = rec.content
                embedding_store.cmetadata = metadata
                result = type(
                    'Result', (), {
                        'EmbeddingStore': embedding_store,
                        'distance': rec.distance
                    }
                )
                formatted_results.append(result)

            return formatted_results

    # aliases for different langchain versions
    def _PGVector__query_collection(self, *args, **kwargs):

        return self.__query_collection(*args, **kwargs)

    def _query_collection(self, *args, **kwargs):
        return self.__query_collection(*args, **kwargs)

    def create_collection(self):
        raise RuntimeError("Forbidden")

    def delete_collection(self):
        raise RuntimeError("Forbidden")

    def delete(self, *args, **kwargs):
        raise RuntimeError("Forbidden")

    def add_embeddings(self, *args, **kwargs):
        raise RuntimeError("Forbidden")



import requests
from datetime import datetime
from typing import Dict, Optional

import pandas as pd
from mlflow.tracking import MlflowClient

from mindsdb.integrations.libs.base import BaseMLEngine


class MLflowHandler(BaseMLEngine):
    """
    The MLflow integration engine needs to have a working connection to MLFlow. For this:
        - All models to use should be previously served
        - An MLflow server should be running, to access its model registry

    Example:
        1. Run `mlflow server -p 5001 --backend-store-uri sqlite:////path/to/mlflow.db --default-artifact-root ./artifacts --host 0.0.0.0`
        2. Run `mlflow models serve --model-uri ./model_path`
        3. Run MindsDB

    Note: above, `artifacts` is a folder to store artifacts for new experiments that do not specify an artifact store.
    """  # noqa

    name = "mlflow"

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:
        args = args["using"]  # ignore the rest of the problem definition
        connection = MlflowClient(args["mlflow_server_url"], args["mlflow_server_path"])
        model_name = args["model_name"]
        mlflow_models = [model.name for model in connection.search_registered_models()]

        if model_name not in mlflow_models:
            raise Exception(
                f"Error: model '{model_name}' not found in mlflow. Check serving and try again."
            )

        args["target"] = target
        self._check_model_url(args["predict_url"])
        self.model_storage.json_set("args", args)

    def predict(self, df, args=None):
        args = self.model_storage.json_get("args")  # override any incoming args for now
        self._check_model_url(args["predict_url"])
        resp = requests.post(
            args["predict_url"],
            data=df.to_json(orient="records"),
            headers={"content-type": "application/json; format=pandas-records"},
        )
        answer = resp.json()
        predictions = pd.DataFrame({args["target"]: answer})
        return predictions

    def describe(self, key: Optional[str] = None) -> pd.DataFrame:
        if key == "info":
            args = self.model_storage.json_get("args")
            connection = MlflowClient(
                args["mlflow_server_url"], args["self.mlflow_server_path"]
            )
            models = {
                model.name: model for model in connection.search_registered_models()
            }
            model = models[key]
            latest_version = model.latest_versions[-1]
            description = {
                "NAME": [model.name],
                "USER_DESCRIPTION": [model.description],
                "LAST_STATUS": [latest_version.status],
                "CREATED_AT": [
                    datetime.fromtimestamp(model.creation_timestamp // 1000).strftime(
                        "%m/%d/%Y, %H:%M:%S"
                    )
                ],
                "LAST_UPDATED": [
                    datetime.fromtimestamp(
                        model.last_updated_timestamp // 1000
                    ).strftime("%m/%d/%Y, %H:%M:%S")
                ],
                "TAGS": [model.tags],
                "LAST_RUN_ID": [latest_version.run_id],
                "LAST_SOURCE_PATH": [latest_version.source],
                "LAST_USER_ID": [latest_version.user_id],
                "LAST_VERSION": [latest_version.version],
            }
            return pd.DataFrame.from_dict(description)
        else:
            tables = ["info"]
            return pd.DataFrame(tables, columns=["tables"])

    @staticmethod
    def _check_model_url(url):
        """try post without data, check status code not in (not_found, method_not_allowed)"""
        try:
            resp = requests.post(url)
            if resp.status_code in (404, 405):
                raise Exception(
                    f"Model url is incorrect, status_code: {resp.status_code}"
                )
        except requests.RequestException as e:
            raise Exception(f"Model url is incorrect: {str(e)}")



import copy
import importlib
from typing import Dict, Union

import pandas as pd
from pandas import DataFrame
from pydantic import BaseModel

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log
from langchain_core.embeddings import Embeddings
from mindsdb.integrations.handlers.langchain_embedding_handler.vllm_embeddings import VLLMEmbeddings
from mindsdb.integrations.handlers.langchain_embedding_handler.fastapi_embeddings import FastAPIEmbeddings

logger = log.getLogger(__name__)

# construct the embedding model name to the class mapping
# we try to import all embedding models from langchain_community.embeddings
# for each class, we get a more user friendly name for it
# E.g. OpenAIEmbeddings -> OpenAI
# This is used for the user to select the embedding model
EMBEDDING_MODELS = {
    'VLLM': 'VLLMEmbeddings',
    'vllm': 'VLLMEmbeddings',
    'FastAPI': 'FastAPIEmbeddings',
    'fastapi': 'FastAPIEmbeddings'

}

try:
    module = importlib.import_module("langchain_community.embeddings")
    # iterate __all__ to get all the classes
    for class_name in module.__all__:
        class_ = getattr(module, class_name)
        if not issubclass(class_, Embeddings):
            continue
        # convert the class name to a more user friendly name
        # e.g. OpenAIEmbeddings -> OpenAI
        user_friendly_name = class_name.replace("Embeddings", "")
        EMBEDDING_MODELS[user_friendly_name] = class_name
        EMBEDDING_MODELS[user_friendly_name.lower()] = class_name

except ImportError:
    raise Exception(
        "The langchain is not installed. Please install it with `pip install langchain-community`."
    )


def get_langchain_class(class_name: str) -> Embeddings:
    """Returns the class object of the handler class.

    Args:
        class_name (str): Name of the class

    Returns:
        langchain.embeddings.BaseEmbedding: The class object
    """
    # First check if it's our custom VLLMEmbeddings
    if class_name == "VLLMEmbeddings":
        return VLLMEmbeddings

    if class_name == "FastAPIEmbeddings":
        return FastAPIEmbeddings

    # Then try langchain_community.embeddings
    try:
        module = importlib.import_module("langchain_community.embeddings")
        class_ = getattr(module, class_name)
    except ImportError:
        raise Exception(
            "The langchain is not installed. Please install it with `pip install langchain`."
        )
    except AttributeError:
        raise Exception(
            f"Could not find the class {class_name} in langchain_community.embeddings. Please check the class name."
        )
    return class_


def construct_model_from_args(args: Dict) -> Embeddings:
    """
    Deserializes the model from the model storage
    """
    target = args.pop("target", None)
    class_name = args.pop("class", LangchainEmbeddingHandler.DEFAULT_EMBEDDING_CLASS)
    if class_name in EMBEDDING_MODELS:
        logger.info(
            f"Mapping the user friendly name {class_name} to the class name: {EMBEDDING_MODELS[class_name]}"
        )
        class_name = EMBEDDING_MODELS[class_name]
    MODEL_CLASS = get_langchain_class(class_name)
    serialized_dict = copy.deepcopy(args)

    # Make sure we don't pass in unnecessary arguments.
    if issubclass(MODEL_CLASS, BaseModel):
        serialized_dict = {
            k: v for k, v in serialized_dict.items() if k in MODEL_CLASS.model_fields
        }

    model = MODEL_CLASS(**serialized_dict)
    if target is not None:
        args["target"] = target
    args["class"] = class_name
    return model


def row_to_document(row: pd.Series) -> str:
    """
    Convert a row in the input dataframe into a document

    Default implementation is to concatenate all the columns
    in the form of
    field1: value1\nfield2: value2\n...
    """
    fields = row.index.tolist()
    values = row.values.tolist()
    document = "\n".join(
        [f"{field}: {value}" for field, value in zip(fields, values)]
    )
    return document


class LangchainEmbeddingHandler(BaseMLEngine):
    """
    Bridge class to connect langchain.embeddings module to mindsDB
    """

    DEFAULT_EMBEDDING_CLASS = "OpenAIEmbeddings"

    def __init__(self, model_storage, engine_storage, **kwargs) -> None:
        super().__init__(model_storage, engine_storage, **kwargs)
        self.generative = True

    def create(
        self,
        target: str,
        df: Union[DataFrame, None] = None,
        args: Union[Dict, None] = None,
    ) -> None:
        # get the class name from the args
        user_args = args.get("using", {})

        # infer the input columns arg if user did not provide it
        # from the columns of the input dataframe if it is provided
        if "input_columns" not in user_args and df is not None:
            # ignore private columns starts with __mindsdb
            # ignore target column in the input dataframe
            user_args["input_columns"] = [
                col
                for col in df.columns.tolist()
                if not col.startswith("__mindsdb") and col != target
            ]
            # unquote the column names -- removing surrounding `
            user_args["input_columns"] = [
                col.strip("`") for col in user_args["input_columns"]
            ]

        elif "input_columns" not in user_args:
            # set as empty list if the input_columns is not provided
            user_args["input_columns"] = []

        # this may raise an exception if
        # the arguments are not sufficient to create such as class
        # due to e.g., lack of API key
        # But the validation logic is handled by langchain and pydantic
        construct_model_from_args(user_args)

        # save the model to the model storage
        target = target or "embeddings"
        user_args[
            "target"
        ] = target  # this is the name of the column to store the embeddings
        self.model_storage.json_set("args", user_args)

    def predict(self, df: DataFrame, args) -> DataFrame:
        # reconstruct the model from the model storage
        user_args = self.model_storage.json_get("args")
        model = construct_model_from_args(user_args)

        # get the target from the model storage
        target = user_args["target"]
        # run the actual embedding vector generation
        # TODO: need a better way to handle this
        # unquote the column names -- removing surrounding `
        cols_dfs = [col.strip("`") for col in df.columns.tolist()]
        df.columns = cols_dfs
        # if input_columns is an empty list, use all the columns
        input_columns = user_args.get("input_columns") or df.columns.tolist()
        # check all the input columns are in the df
        if not all(
            # ignore surrounding ` in the column names when checking
            [col in cols_dfs for col in input_columns]
        ):
            raise Exception(
                f"Input columns {input_columns} not found in the input dataframe. Available columns are {df.columns}"
            )

        # convert each row into a document
        df_texts = df[input_columns].apply(row_to_document, axis=1)
        embeddings = model.embed_documents(df_texts.tolist())

        # create a new dataframe with the embeddings
        df_embeddings = df.copy().assign(**{target: embeddings})

        return df_embeddings

    def finetune(
        self, df: Union[DataFrame, None] = None, args: Union[Dict, None] = None
    ) -> None:
        raise NotImplementedError(
            "Finetuning is not supported for langchain embeddings"
        )

    def describe(self, attribute: Union[str, None] = None) -> DataFrame:
        args = self.model_storage.json_get("args")

        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])
        elif attribute == "metadata":
            return pd.DataFrame(
                [
                    ("model_class", self.model_storage.json_get("model_class")),
                ],
                columns=["key", "value"],
            )

        else:
            tables = ("args", "metadata")
            return pd.DataFrame(tables, columns=["tables"])



from typing import List


class TextSplitter:
    def __init__(
        self,
        chunk_size: int = 1000,
        chunk_overlap: int = 200,
        separators: List[str] = None,
        k_range: float = 0.5,
        k_ratio: float = 1,
    ):
        """
        Split text into chunks. The logic:
         - Get a piece of text with chunk_size and try to find the separator at the end of the piece.
         - The allowed range to find the separator is defined by k_range and k_ratio using formula:
            k_range * chunk_size / (num * k_ratio + 1)
            num - is number of a separator from the list
         - if the separator is not in the rage: switch to the next separator
         - if the found separator is in the middle of the sentence, use overlapping:
            - the found text is the current chunk
            - repeat the search with less strict k_range and k_ratio
            - the found text will be the beginning of the next chunk

        :param chunk_size: size of the chunk, which must not be exceeded
        :param separators: list of separators in order of priority
        :param k_range: defines the range to look for the separator
        :param k_ratio: defines how much to shrink the range for the next separator
        """
        if separators is None:
            separators = ["\n\n", "\n", ". ", " ", ""]
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = separators
        self.k_range = k_range
        self.k_ratio = k_ratio

    def split_text(self, text: str) -> List[str]:
        chunks = []

        while True:
            if len(text) < self.chunk_size:
                chunks.append(text)
                break

            sep, chunk, shift = self.get_next_chunk(text, self.k_range, self.k_ratio)
            chunks.append(chunk)

            text = text[shift:]
        return chunks

    def get_next_chunk(self, text: str, k_range: float, k_ratio: float):
        # returns chunk with separator and shift for the next search iteration

        chunk = text[: self.chunk_size]
        # positions = []
        for i, sep in enumerate(self.separators):
            pos = chunk.rfind(sep)

            vpos = self.chunk_size - pos
            if vpos < k_range * self.chunk_size / (i * k_ratio + 1):
                shift = len(sep) + pos
                if sep.strip(" ") == "":
                    # overlapping
                    sep2, _, shift2 = self.get_next_chunk(text, k_range * 1.5, 0)
                    if sep2.strip(" ") != "":
                        # use shift of previous separator
                        if shift - shift2 < self.chunk_overlap:
                            shift = shift2

                return sep, chunk[:pos], shift

        raise RuntimeError("Cannot split text")



import unittest
from mindsdb.integrations.handlers.vertica_handler.vertica_handler import VerticaHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class VerticaHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": '127.0.0.1',
                "port": 5433,
                "user": 'dbadmin',
                "password": '',
                "database": 'VMart',
                "schema_name": 'public'
            }

        }
        cls.handler = VerticaHandler('test_vertica_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_connect(self):
        assert self.handler.connect()

    def test_2_create_table(self):
        query = "CREATE Table TEST(id Number(1),Name Varchar(33))"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR

    def test_3_insert(self):
        query = "INSERT INTO TEST (1,'lOVe yOU)"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR

    def test_4_native_query_select(self):
        query = "SELECT * FROM TEST;"
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        columns = self.handler.get_columns('TEMP')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



"""project-company

Revision ID: c06c35f7e8e1
Revises: f6dc924079fa
Create Date: 2025-01-15 14:14:29.295834

"""
from collections import defaultdict

from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa
from mindsdb.utilities import log

# revision identifiers, used by Alembic.
revision = 'c06c35f7e8e1'
down_revision = 'f6dc924079fa'
branch_labels = None
depends_on = None


logger = log.getLogger(__name__)


def upgrade():

    """
    convert company_id from null to 0 to make constrain works
    duplicated names are renamed
    """

    conn = op.get_bind()
    table = sa.Table(
        'project',
        sa.MetaData(),
        sa.Column('id', sa.Integer()),
        sa.Column('name', sa.String()),
        sa.Column('company_id', sa.Integer()),
    )

    data = conn.execute(
        table
        .select()
        .where(table.c.company_id == sa.null())
    ).fetchall()

    names = defaultdict(list)
    for id, name, _ in data:
        names[name].append(id)

    # get duplicated
    for name, ids in names.items():
        if len(ids) == 1:
            continue

        # rename all except first
        for id in ids[1:]:
            new_name = f'{name}__{id}'

            op.execute(
                table
                .update()
                .where(table.c.id == id)
                .values({'name': new_name})
            )
            logger.warning(f'Found duplicated project name: {name}, renamed to: {new_name}')

    op.execute(
        table
        .update()
        .where(table.c.company_id == sa.null())
        .values({'company_id': 0})
    )


def downgrade():
    table = sa.Table(
        'project',
        sa.MetaData(),
        sa.Column('company_id', sa.Integer())
    )

    op.execute(
        table
        .update()
        .where(table.c.company_id == 0)
        .values({'company_id': sa.null()})
    )
